/*
Eos Testbed
2013 - Matt Pierce

	Creates an SDL window, creates a scene in Eos,
	renders scene. This barebones app is intended to
	ensure that Eos compiles correctly and that its
	basics are functioning properly.

TAB WIDTH: 4
*/

#include <stdio.h>						// standard C header
#include <math.h>						// standard C math functions
#include <time.h>						// for srand(time(NULL))
#include "SDL.h"						// SDL library
#include "SDL_image.h"					// SDL Image library (for image file IO)
#include "SDL_ttf.h"					// SDL TrueType library (for putting text on the screen)

#include "eos/eos.h"					// EOS Raytracer

#include "gameStates/sampleScene.h"		// Sample Scene

// GLOBAL CONSTANTS
const int		InternalWidth = 640;
const int		InternalHeight = 480;
const int		DefaultScreenWidth = 640;		// The width of our SDL window
const int		DefaultScreenHeight = 480;		// The height of our SDL window
const bool		FullScreen = false;		// whether or not we should go full screen
const float		pi=3.1415926535897932384626433832795f; // PI
const int		nRendererStates = 6;
unsigned const int		rendererStates[nRendererStates] = {
	EOS_RENDERER_SUBSAMPLE|EOS_RENDERER_MULTITHREADED,
	EOS_RENDERER_SUBSAMPLE|EOS_RENDERER_MULTITHREADED|EOS_RENDERER_SUBSAMPLE_DIFFUSE,
	EOS_RENDERER_SUBSAMPLE|EOS_RENDERER_MULTITHREADED|EOS_RENDERER_SUBSAMPLE_DIFFUSE|EOS_RENDERER_SUBSAMPLE_SPECULAR,
	EOS_RENDERER_SUBSAMPLE|EOS_RENDERER_MULTITHREADED|EOS_RENDERER_SUBSAMPLE_DIFFUSE|EOS_RENDERER_SUBSAMPLE_SPECULAR|EOS_RENDERER_SUBSAMPLE_REFLECTION,
	EOS_RENDERER_STANDARD|EOS_RENDERER_JITTER_SOFT_SHADOWS|EOS_RENDERER_MULTITHREADED,
	EOS_RENDERER_STANDARD|EOS_RENDERER_JITTER_SOFT_SHADOWS|EOS_RENDERER_MULTITHREADED|EOS_RENDERER_ANTIALIAS
};

// GLOBAL VARIABLES

// SCENE AND CAMERA
eosRendererSettings rendererSettings;
const float		PlayerMouseSensitivity = 0.005f;
const float		PlayerKeyboardSensitivity = 0.04f;
GSSampleScene	mySampleScene;

// GLOBAL FUNCTIONS
void InitGame(SampleSceneInitData *initData);
void FinishOff(SampleSceneInitData *initData);

/// <summary>The start of our program, runs the main event message loop for SDL.</summary>
/// <returns>0</returns>
int main(int argc, char* argv[])
{
	int				gameRunning = 1;			// variable for ending our event loop
	Uint32			lastFrame = SDL_GetTicks();
	Uint32			currentFrame;
	float			dt;

	SampleSceneInitData initData;
	initData.gameRunning = &gameRunning;
	InitGame(&initData);									// Init SDL and allocate storage

	mySampleScene.Load();
	mySampleScene.Init((void *)(&initData));

	while (gameRunning) {
		currentFrame = SDL_GetTicks();
		dt = (float)(currentFrame-lastFrame)/1000.0f;

		mySampleScene.Cycle(dt, NULL);
		mySampleScene.Draw();

		lastFrame = currentFrame;
	}

	mySampleScene.Unload();

	FinishOff(&initData);								// free storage and quit SDL
	return 0;
}

/// <summary>Initializes SDL Library and allocates storage</summary>
/// <returns>(void)</returns>
void InitGame(SampleSceneInitData *initData) {
	// init rand
	srand((unsigned int)time(NULL));

	initData->DefaultScreenWidth = DefaultScreenWidth;
	initData->DefaultScreenHeight = DefaultScreenHeight;
	initData->InternalWidth = InternalWidth;
	initData->InternalHeight = InternalHeight;
	initData->PlayerKeyboardSensitivity = PlayerKeyboardSensitivity;
	initData->PlayerMouseSensitivity = PlayerMouseSensitivity;
	initData->rendererSettings = &rendererSettings;
	initData->FullScreen = FullScreen;

	// init SDL
	SDL_Init(SDL_INIT_EVERYTHING);
	SDL_WM_GrabInput(SDL_GRAB_ON);
	SDL_ShowCursor(SDL_FALSE);
	TTF_Init();
	initData->screen = SDL_SetVideoMode(FullScreen?0:DefaultScreenWidth, FullScreen?0:DefaultScreenHeight, 32, SDL_HWSURFACE | SDL_DOUBLEBUF | (FullScreen?SDL_FULLSCREEN:0));
	SDL_WM_SetCaption("SDL Test", NULL);

	// renderer settings
	rendererSettings.fogRate = 0.0004f;						// absorb 10% light for each 1 unit
	rendererSettings.fogColor[0] = 0.40f*0.4f;
	rendererSettings.fogColor[1] = 0.60f*0.4f;
	rendererSettings.fogColor[2] = 0.68f*0.4f;
	rendererSettings.subdivisionThreshold = 0.15f;
	rendererSettings.reflectionThreshold = 0.05f;
	rendererSettings.antialiasThreshold = 0.5f;
	rendererSettings.ambientOcclusionDistance = 20.0f;
}

/// <summary>Shuts down SDL Library and frees storage</summary>
/// <returns>(void)</returns>
void FinishOff(SampleSceneInitData *initData) {
	TTF_Quit();
    SDL_Quit();
    exit(0);
}
