/*
Eos Filters - definitions
2013 - Matt Pierce

	Defines the functions declared in eosFilters.h

TAB WIDTH: 4
*/

#include <math.h>
#include "eosFilters.h"
#include "SDL.h"						// SDL library

/// <summary>Simulates camera exposure on an RGB float buffer</summary>
/// <returns>(void)</returns>
void eosFilterExposure(float exposure, float *buff, unsigned int w, unsigned int h) {
	unsigned int i, j;
	float *tBuff = buff;
	const float e = 2.71828182845904523536f;

	for (j=0;j<h;j++) {
		for (i=0;i<w;i++) {
			tBuff[0] = 1.0f - powf(e, -tBuff[0]*exposure);
			tBuff[1] = 1.0f - powf(e, -tBuff[1]*exposure);
			tBuff[2] = 1.0f - powf(e, -tBuff[2]*exposure);
			tBuff += 3;
		}
	}
}

/// <summary>Gamma corrects an RGB float buffer</summary>
/// <returns>(void)</returns>
void eosFilterGamma(float gamma, float *buff, unsigned int w, unsigned int h) {
	unsigned int i, j;
	float *tBuff = buff;

	gamma = 1.0f/gamma;

	for (j=0;j<h;j++) {
		for (i=0;i<w;i++) {
			tBuff[0] = powf(tBuff[0], gamma);
			tBuff[1] = powf(tBuff[1], gamma);
			tBuff[2] = powf(tBuff[2], gamma);
			tBuff += 3;
		}
	}
}

/// <summary>Fast 24-bit exposure estimation.</summary>
/// <returns>(void)</returns>
void eosFilterFastExposure(float exposure, unsigned char *buff, unsigned int w, unsigned int h) {
	unsigned int i, j;
	unsigned char *tBuff = buff;
	const float e = 2.71828182845904523536f;
	unsigned char exposureTable[256];

	for (i=0;i<256;i++) { 
		j = (unsigned int)floor(256.0f * (1.0f - powf(e, -(i/255.0f)*exposure)));
		exposureTable[i] = (j<0?0:j>255?255:j);
	}

	for (j=0;j<h;j++) {
		for (i=0;i<w;i++) {
			tBuff[0] = exposureTable[tBuff[0]];
			tBuff[1] = exposureTable[tBuff[1]];
			tBuff[2] = exposureTable[tBuff[2]];
			tBuff += 3;
		}
	}
}

/// <summary>Fast 24-bit gamma correction</summary>
/// <returns>(void)</returns>
void eosFilterFastGamma(float gamma, unsigned char *buff, unsigned int w, unsigned int h) {
	unsigned int i, j;
	unsigned char *tBuff = buff;
	unsigned char gammaTable[256];

	gamma = 1.0f/gamma;

	for (i=0;i<256;i++) {
		j = (int)(255.0f * powf(i/255.0f, gamma));
		gammaTable[i] = (j<0?0:j>255?255:j);
	}

	for (j=0;j<h;j++) {
		for (i=0;i<w;i++) {
			tBuff[0] = gammaTable[tBuff[0]];
			tBuff[1] = gammaTable[tBuff[1]];
			tBuff[2] = gammaTable[tBuff[2]];
			tBuff += 3;
		}
	}
}

/// <summary>RGB float gamma correction via piecewise linear estimation.
/// For values outside the precalculated range, uses regular slow method.</summary>
/// <returns>(void)</returns>
void eosFilterFastFloatGamma(float gamma, float *buff, unsigned int w, unsigned int h) {
	const int gammaTableSize = 256;
	unsigned int i, k;
	float *tBuff = buff;
	float gammaTable[gammaTableSize];	// will be for gamma (0,1]
	int tabIndex;
	float pct;

	gamma = 1.0f/gamma;

	// populate the table
	for (i=0;i<gammaTableSize;i++) {
		gammaTable[i] = powf((i+1)/(float)gammaTableSize, gamma);
	}

	k = w*h*3;
	for (i=0;i<k;i++) {
		if (*tBuff==0.0f) {
			// do nothing
		} else if (*tBuff<1.0f) {
			// in range, use table
			pct = (*tBuff)*gammaTableSize-1;
			tabIndex = (int)pct;
			pct = pct - tabIndex;
			if (tabIndex<0) tabIndex=0;
			*tBuff = gammaTable[tabIndex]*(1.0f-pct) + gammaTable[tabIndex+1]*pct;
		} else {
			*tBuff = powf(*tBuff, gamma);
		}
		tBuff++;
	}
}

/// <summary>Simulates camera exposure using piecewise linear estimation.
/// For values outside of precalculated range, uses regular slow method.</summary>
/// <returns>(void)</returns>
void eosFilterFastFloatExposure(float exposure, float *buff, unsigned int w, unsigned int h) {
	const int exposureTableSize = 256;
	unsigned int i, k;
	float *tBuff = buff;
	const float e = 2.71828182845904523536f;
	float exposureTable[exposureTableSize];	// will be for gamma (0,1]
	int tabIndex;
	float pct;

	// populate the table
	for (i=0;i<exposureTableSize;i++) {
		exposureTable[i] = 1.0f - powf(e, -((float)i/(float)exposureTableSize)*exposure);
	}

	k = w*h*3;
	for (i=0;i<k;i++) {
		if (*tBuff==0.0f) {
			// do nothing
		} else if (*tBuff<1.0f) {
			// in range, use table
			pct = (*tBuff)*exposureTableSize-1;
			tabIndex = (int)pct;
			pct = pct - tabIndex;
			if (tabIndex<0) tabIndex=0;
			*tBuff = exposureTable[tabIndex]*(1.0f-pct) + exposureTable[tabIndex+1]*pct;
		} else {
			*tBuff = 1.0f - powf(e, -(*tBuff)*exposure);
		}
		tBuff++;
	}
}

void eosFilterSoften(unsigned int iterations, float strength, float *buff, unsigned int w, unsigned int h) {
	unsigned int i, x, y;
	int rowLen = w*3;
	float * src = buff;
	float * dest;
	float is = 1.0f-strength;
	strength *= 0.25f;

	float * srcOffs;
	float * destOffs;

	dest = (float *)malloc(w*h*3*sizeof(float));

	for (i=0;i<iterations;i++) {
		for (y=1;y<h-1;y++) {
			srcOffs = src + 3*(y*w + 1);
			destOffs = dest + (srcOffs-src);
			for (x=1;x<w-1;x++) {
				*destOffs = strength*(*(srcOffs-rowLen) + *(srcOffs+rowLen) + *(srcOffs-3) + *(srcOffs+3)) + is*((*srcOffs));
				++destOffs;
				++srcOffs;
				*destOffs = strength*(*(srcOffs-rowLen) + *(srcOffs+rowLen) + *(srcOffs-3) + *(srcOffs+3)) + is*((*srcOffs));
				++destOffs;
				++srcOffs;
				*destOffs = strength*(*(srcOffs-rowLen) + *(srcOffs+rowLen) + *(srcOffs-3) + *(srcOffs+3)) + is*((*srcOffs));
				++destOffs;
				++srcOffs;
			}
		}
		// swap buffers for next iteration (using srcOffs as temp variable)
		srcOffs = src;
		src = dest;
		dest = srcOffs;
	}

	// memory copy if needed
	if (src == buff) {
		free(dest);
	} else {
		memcpy(buff, src, 3*w*h*sizeof(float));
		free(src);
	}
}

void eosInitExposureTracker(eosExposureTracker *tracker) {
	tracker->minExposure = 0.5f;
	tracker->maxExposure = 5.0f;
	tracker->lastSample = (float)SDL_GetTicks();
	tracker->lastExposure = tracker->targetExposure = tracker->currentExposure = 1.0f;
	tracker->sampleInterval = 1000.0f; // milliseconds
	tracker->targetBrightness = 0.6f;
}


void eosSampleExposure(float *buff, unsigned int w, unsigned int h, eosExposureTracker *tracker) {
	float tNow = (float)SDL_GetTicks();
	if ((tNow-tracker->lastSample)>tracker->sampleInterval) {
		// jump to end of interval
		tracker->lastExposure = tracker->currentExposure = tracker->targetExposure;
		tracker->lastSample = tNow;

		float avg = 0.0f;
		int x, y;

		// Like a camera, we want to meter around the center of the image. This will
		// avoid problems where a light off to the side (like the sun) darkens everything.
		int startX = w >> 2;
		int startY = h >> 2;
		int endX = startX + (w >> 1);
		int endY = startY + (h >> 1);
		int nTaps = 0;
		float *tBuff;
		for (y=startY;y<endY;y++) {
			for (x=startX;x<endX;x++) {
				tBuff = buff + 3*(y*w + x);
				avg += tBuff[0] + tBuff[1] + tBuff[2];
				++nTaps;
			}
		}


		avg = avg / (float)nTaps;
		tracker->targetExposure = -log(1.0f-tracker->targetBrightness)/avg;
		if (tracker->targetExposure < tracker->minExposure) tracker->targetExposure = tracker->minExposure;
		if (tracker->targetExposure > tracker->maxExposure) tracker->targetExposure = tracker->maxExposure;
	} else {
		float t = (tNow - tracker->lastSample)/(tracker->sampleInterval);
		tracker->currentExposure = t*tracker->targetExposure + (1.0f-t)*tracker->lastExposure;
	}
}
