/*
Eos Renderer - definitions
2013 - Matt Pierce

	Defines functions that render the scene.  It's important to make as much of this
	thread-safe as humanly possible so that you can call functions from the
	platform-specific multithreaded renderers instead of duplicating all that
	effort.

	NOTE: There are a few structs defined here instead of in the header file. That's
	done intentionally to hide a little of the complexity from the user. The user
	should only ever need to know how to use the eosRenderScene function and then
	the renderer should take care of everything else automagically.

TAB WIDTH: 4
*/
#include <random>
#include <string.h>				// for memcpy
#include "eosRenderer.h"
#include "eosScene.h"
#include "eosCamera.h"
#include "eosEnvironmentMap.h"
#include "eosLight.h"
#include "VectorMath.h"

#ifdef WIN32
// If we're compiling in Win32 we will need these headers for multithreading
#include <Windows.h>			// For basic Windows types (i.e. DWORD, HANDLE)
#include <process.h>			// For Windows multithreading functions
#endif

// Ambient occlusion randomization
const int ao_rand_size = 0x7FF+1;
float ao_rand_table[ao_rand_size];
int ao_rand_ptr = 0;
bool ao_init = false;

void eosInitAORandTable() {
	int i;
	for(i=0;i<ao_rand_size;i++) {
		ao_rand_table[i] = rand()/(float)RAND_MAX;
	}
	ao_init = true;
}


/*
	+------------------+--------------------+
	| SINGLETHREADED   | STANDARD           |
	+------------------+--------------------+
*/


/// <summary>Blends colors in accordance with the material. This is done so
/// frequently in both standard and subsample renderers, but at different points,
/// so it made sense to make this its own function.</summary>
/// <returns>(void)</returns>
void eosRendererBlendColors(const eosMaterial *mat, const float *surfaceCol, const float *diffuseCol, const float *specularCol, const float *colorBehind, 
							const float *colorReflect, const float fresnelTerm, float fogPct, const float *fogColor, float *output) {
	output[0] = surfaceCol[0]*diffuseCol[0];
	output[1] = surfaceCol[1]*diffuseCol[1];
	output[2] = surfaceCol[2]*diffuseCol[2];

	// specular reflection
	if (mat->specular>0.0f) {
		output[0] += specularCol[0];
		output[1] += specularCol[1];
		output[2] += specularCol[2];
	}

	// gloss reflection
	if (mat->glossReflection) {
		output[0] += mat->glossReflection*colorReflect[0]*fresnelTerm;
		output[1] += mat->glossReflection*colorReflect[1]*fresnelTerm;
		output[2] += mat->glossReflection*colorReflect[2]*fresnelTerm;
	}

	// metalic reflection
	if (mat->metalicReflection) {
		output[0] += mat->metalicReflection*surfaceCol[0]*colorReflect[0]*fresnelTerm;
		output[1] += mat->metalicReflection*surfaceCol[1]*colorReflect[1]*fresnelTerm;
		output[2] += mat->metalicReflection*surfaceCol[2]*colorReflect[2]*fresnelTerm;
	}

	if (surfaceCol[3]<1.0f) {
		float a = surfaceCol[3];
		output[0] = a*output[0] + (1.0f-a)*colorBehind[0];
		output[1] = a*output[1] + (1.0f-a)*colorBehind[1];
		output[2] = a*output[2] + (1.0f-a)*colorBehind[2];
	}

	if (fogPct > 0.0f) {
		output[0] = (1.0f-fogPct)*output[0] + fogPct*fogColor[0];
		output[1] = (1.0f-fogPct)*output[1] + fogPct*fogColor[1];
		output[2] = (1.0f-fogPct)*output[2] + fogPct*fogColor[2];
	}
}

/// <summary>Starting from surface point P and direction D toward light, finds out
/// how much light is occluded between here and there. Accounts for translucency
/// and surface color on translucent materials.</summary>
/// <returns>(void)</returns>
void eosRendererGetShadowColor(eosScene *scene, const float &px, const float &py, const float &pz, const float &dx, const float &dy, const float &dz, float capT, float *returnColor, float *returnT=NULL) {
	eosTraceResult tr;
	
	tr = eosTraceScene(scene, px, py, pz, dx, dy, dz, capT);
	if (tr.t>0) {
		eosMaterial *			mat;			// pointer to material
		eosTriangle *			tri;			// pointer to triangle
		eosTriangleRenderInfo *	renderInfo;		// pointer to render info
		SVect3d					tcContact;		// texcoords at contact
		float					surfaceCol[4];	// surface color

		if (returnT != NULL) *returnT = tr.t;

		// get pointers to our triangle, render info, and material
		tri = &(scene->meshes[tr.mesh]->tris[tr.tri]);
		renderInfo = &(scene->meshes[tr.mesh]->renderInfo[tr.tri]);
		mat = renderInfo->mat;

		// get surface color
		if (mat->tex  && mat->castsShadow) {
			// if this map is able to have translucency, get the texel
			if (mat->tex->planes == 2 || mat->tex->planes == 4) {
				tcContact = tr.a1*v3d(renderInfo->tc1[0], renderInfo->tc1[1], renderInfo->tc1[2]) +
							tr.a2*v3d(renderInfo->tc2[0], renderInfo->tc2[1], renderInfo->tc2[2]) +
							tr.a3*v3d(renderInfo->tc3[0], renderInfo->tc3[1], renderInfo->tc3[2]);
				eosSampleTexture(*(mat->tex), tcContact.x, tcContact.y, surfaceCol);
			} else {
				// otherwise, just say opaque, color not important
				surfaceCol[3] = 1.0f;
			}
		} else {
			memcpy(surfaceCol, mat->color, 4*sizeof(float));
		}

		// if the surface is totally opaqe, return black
		if (surfaceCol[3] >= 1.0f && mat->castsShadow) {
			returnColor[0] = returnColor[1] = returnColor[2] = 0.0f;
		} else {
			if (mat->castsShadow) {
				// if this thing casts a shadow, get the surface color, light behind
				// and return modulated result
				float tempCol[3];
				if (capT) {
					SVect3d lightPos = v3d(px,py,pz)+v3d(dx,dy,dz);
					SVect3d p = v3d(px, py, pz) + tr.t*v3d(dx,dy,dz) + 0.001f*vmNormalize(v3d(dx,dy,dz));
					SVect3d d = lightPos-p;
					eosRendererGetShadowColor(scene, p.x, p.y, p.z, d.x, d.y, d.z, 1.0f, tempCol);
				} else {
					SVect3d p = v3d(px, py, pz) + tr.t*v3d(dx,dy,dz) + 0.001f*vmNormalize(v3d(dx,dy,dz));
					eosRendererGetShadowColor(scene, p.x, p.y, p.z, dx, dy, dz, -1.0f, tempCol);
				}
				returnColor[0] = tempCol[0]*(surfaceCol[0])*(1.0f-surfaceCol[3]);
				returnColor[1] = tempCol[1]*(surfaceCol[1])*(1.0f-surfaceCol[3]);
				returnColor[2] = tempCol[2]*(surfaceCol[2])*(1.0f-surfaceCol[3]);
			} else {
				// if it doesn't cast a shadow, just return the color behind
				if (capT) {
					SVect3d lightPos = v3d(px,py,pz)+v3d(dx,dy,dz);
					SVect3d p = v3d(px, py, pz) + tr.t*v3d(dx,dy,dz) + 0.001f*vmNormalize(v3d(dx,dy,dz));
					SVect3d d = lightPos-p;
					eosRendererGetShadowColor(scene, p.x, p.y, p.z, d.x, d.y, d.z, 1.0f, returnColor);
				} else {
					SVect3d p = v3d(px, py, pz) + tr.t*v3d(dx,dy,dz) + 0.001f*vmNormalize(v3d(dx,dy,dz));
					eosRendererGetShadowColor(scene, p.x, p.y, p.z, dx, dy, dz, -1.0f, returnColor);
				}
			}
		}
	} else {
		// didn't hit anything, so full opacity
		returnColor[0] = returnColor[1] = returnColor[2] = 1.0f;
	}
}

void eosRendererInterpolateSoftShadow(eosScene *scene, SVect3d p, SVect3d d1, SVect3d d2, SVect3d d3, float *col1, float *col2, float *col3, float *outputCol, const eosRendererSettings &settings, int levels=2) {
	float diff = abs((col1[0]+col1[1]+col1[2]) - (col2[0]+col2[1]+col2[2]));
	diff += abs((col3[0]+col3[1]+col3[2]) - (col2[0]+col2[1]+col2[2]));
	diff += abs((col3[0]+col3[1]+col3[2]) - (col1[0]+col1[1]+col1[2]));

	if ((levels<=0) || (diff<settings.softLightThreshold)) {
		outputCol[0] = (col1[0]+col2[0]+col3[0])/3.0f;
		outputCol[1] = (col1[1]+col2[1]+col3[1])/3.0f;
		outputCol[2] = (col1[2]+col2[2]+col3[2])/3.0f;
	} else {
		SVect3d d12, d23, d31;
		float col12[3], col23[3], col31[3];
		float col_a[3], col_b[3], col_c[3], col_d[3];

		d12 = 0.5f*(d1+d2);
		d23 = 0.5f*(d2+d3);
		d31 = 0.5f*(d3+d1);

		eosRendererGetShadowColor(scene, p.x, p.y, p.z, d12.x, d12.y, d12.z, -1.0f, col12);
		eosRendererGetShadowColor(scene, p.x, p.y, p.z, d23.x, d23.y, d23.z, -1.0f, col23);
		eosRendererGetShadowColor(scene, p.x, p.y, p.z, d31.x, d31.y, d31.z, -1.0f, col31);

		eosRendererInterpolateSoftShadow(scene, p, d1, d12, d31, col1, col12, col31, col_a, settings, levels-1);
		eosRendererInterpolateSoftShadow(scene, p, d12, d2, d23, col12, col2, col23, col_b, settings, levels-1);
		eosRendererInterpolateSoftShadow(scene, p, d31, d23, d3, col31, col23, col3, col_c, settings, levels-1);
		eosRendererInterpolateSoftShadow(scene, p, d12, d23, d31, col12, col23, col31, col_d, settings, levels-1);

		outputCol[0] = 0.25f*(col_a[0] + col_b[0] + col_c[0] + col_d[0]);
		outputCol[1] = 0.25f*(col_a[1] + col_b[1] + col_c[1] + col_d[1]);
		outputCol[2] = 0.25f*(col_a[2] + col_b[2] + col_c[2] + col_d[2]);
	}
}

void eosRendererGetSoftShadowColor(eosScene *scene, const SVect3d &p, const eosLight &l, float *outputCol, const eosRendererSettings &settings) {
	float col1[3], col2[3], col3[3];

	if (settings.flags & EOS_RENDERER_JITTER_SOFT_SHADOWS) {
		// init thread-safe random number generator
		std::uniform_real_distribution<> dis(0, 1);
		std::random_device rd;
		std::mt19937 gen(rd());
	
		CMatrix tMatrix;
		tMatrix.LoadIdentity();
		tMatrix.Rotate(6.283185307179586476925286766559f*dis(rd), l.d[0], l.d[1], l.d[2]);

		SVect3d ld2 = tMatrix.Transform(v3d(l.d2[0], l.d2[1], l.d2[2]));
		SVect3d ld3 = tMatrix.Transform(v3d(l.d3[0], l.d3[1], l.d3[2]));
		SVect3d ld4 = tMatrix.Transform(v3d(l.d4[0], l.d4[1], l.d4[2]));
	
		eosRendererGetShadowColor(scene, p.x, p.y, p.z, ld2.x, ld2.y, ld2.z, -1.0f, col1);
		eosRendererGetShadowColor(scene, p.x, p.y, p.z, ld3.x, ld3.y, ld3.z, -1.0f, col2);
		eosRendererGetShadowColor(scene, p.x, p.y, p.z, ld4.x, ld4.y, ld4.z, -1.0f, col3);

		eosRendererInterpolateSoftShadow(scene, p, ld2, ld3, ld4, col1, col2, col3,outputCol,settings);
	} else {
		SVect3d ld2 = (v3d(l.d2[0], l.d2[1], l.d2[2]));
		SVect3d ld3 = (v3d(l.d3[0], l.d3[1], l.d3[2]));
		SVect3d ld4 = (v3d(l.d4[0], l.d4[1], l.d4[2]));

		eosRendererGetShadowColor(scene, p.x, p.y, p.z, ld2.x, ld2.y, ld2.z, -1.0f, col1);
		eosRendererGetShadowColor(scene, p.x, p.y, p.z, ld3.x, ld3.y, ld3.z, -1.0f, col2);
		eosRendererGetShadowColor(scene, p.x, p.y, p.z, ld4.x, ld4.y, ld4.z, -1.0f, col3);

		eosRendererInterpolateSoftShadow(scene, p, ld2, ld3, ld4, col1, col2, col3,outputCol,settings);
	}
}

/// <summary>Finds the ambient occlusion on a surface</summary>
/// <returns>(void)</returns>
void eosRendererGetAOValue(eosScene *scene, SVect3d p, SVect3d n, float *outputCol, const eosRendererSettings &settings) {
	int				i;				// loop counter
	float			potential=0.0f;	// maximum amount of light we could've gathered
	float			dp;				// dot product for current trace
	float			tmpCol[3];		// returned color
	float			u, v;			// our random values
	SVect3d			d;				// direction to search
	float			dist;			// distance to first hit

	outputCol[0] = outputCol[1] = outputCol[2] = 0.0f;

	for (i=0;i<settings.ambientOcclusionSamples;i++) {
		u = ao_rand_table[(ao_rand_ptr++)&(ao_rand_size-1)];
		v = ao_rand_table[(ao_rand_ptr++)&(ao_rand_size-1)];

		d.z = v;

		u = u*6.283185307179586476925286766559f;
		v = sqrtf(1.0f-v*v);

		d.x = v*cos(u);
		d.y = v*sin(u);

		dp = vmDotProduct(d, n);
		if (dp<0) {
			d = -d;
			dp = -dp;
		}

		dist = -1;
		eosRendererGetShadowColor(scene, p.x, p.y, p.z, d.x, d.y, d.z, settings.ambientOcclusionDistance, tmpCol, &dist);

		if ((dist > 0)  && (settings.ambientOcclusionDistance > 0)) {
			dist = vmDistance(d*dist, v3d(0.0f, 0.0f, 0.0f));
			dist = dist/settings.ambientOcclusionDistance;
			if (dist < 0.0f) {
				dist = 0.0f;
			}
		} else {
			dist = 0.0f;
		}
		if (dist > 1.0f) dist = 1.0f;

		outputCol[0] += dp*(tmpCol[0]*(1.0f-dist) + dist);
		outputCol[1] += dp*(tmpCol[1]*(1.0f-dist) + dist);
		outputCol[2] += dp*(tmpCol[2]*(1.0f-dist) + dist);
		potential += dp;
	}

	outputCol[0] /= potential;
	outputCol[1] /= potential;
	outputCol[2] /= potential;
}

/// <summary>Traces ray P+Dt and boils it down to the color that should be displayed. This is done
/// obviously once for every pixel in the standard renderer, but also handles reflection colors and
/// colors behind translucent objects in both renderers</summary>
/// <returns>(void)</returns>
void eosRendererTraceColor(eosScene *scene, const float &px, const float &py, const float &pz, const float &dx, const float &dy, const float &dz, float *returnColor, int level, const eosRendererSettings &settings) {
	eosTraceResult tr;

	if (level == 0) {
		// recursion limit, return black
		returnColor[0] = returnColor[1] = returnColor[2] = 0.0f;
	} else {

		tr = eosTraceScene(scene, px, py, pz, dx, dy, dz, -1.0f);
		if (tr.t>0) {
			int						i;					// loop counters
			SVect3d					pContact;			// point of contact
			SVect3d					nContact;			// normal at contact
			SVect3d					nPlane;				// normal of the plane itself
			SVect3d					tcContact;			// texcoords at contact
			SVect3d					nd;					// negative trace direction (used more frequently than positive)
			SVect3d					rVect;				// reflection vector
			float					fresnelTerm;		// fresnel term
			float					surfaceCol[4];		// surface color
			eosMaterial *			mat;				// pointer to material
			eosTriangle *			tri;				// pointer to triangle
			eosTriangleRenderInfo *	renderInfo;			// pointer to render info
			float					diffuseCol[4];		// diffuse light color
			float					specularCol[3];		// specular light color
			float					colorBehind[3];		// color behind this contact (if alpha<1.0)
			float					colorReflect[3];	// color from reflection
			float					fogPct=0.0f;		// fog blending value
			bool					backSide = false;
			bool					needsAO = false;
			float					AOColor[3];			// ambient occlusion multiplier

			nd = -vmNormalize(v3d(dx, dy, dz));

			// get pointers to our triangle, render info, and material
			tri = &(scene->meshes[tr.mesh]->tris[tr.tri]);
			renderInfo = &(scene->meshes[tr.mesh]->renderInfo[tr.tri]);
			mat = renderInfo->mat;

			// get the plane normal in world space
			nPlane = v3d(tri->pnx, tri->pny, tri->pnz);
			nPlane = scene->meshes[tr.mesh]->tree->matrix.NormalTransform(nPlane);
			nPlane = scene->meshes[tr.mesh]->matrix.NormalTransform(nPlane);

			// use trace result to calcuate contact point and normal at contact
			/*
			pContact =	tr.a1*v3d(tri->p1[0], tri->p1[1], tri->p1[2]) +
						tr.a2*v3d(tri->p2[0], tri->p2[1], tri->p2[2]) +
						tr.a3*v3d(tri->p3[0], tri->p3[1], tri->p3[2]);
			*/
			pContact = v3d(px + 0.999f*tr.t*dx, py + 0.999f*tr.t*dy, pz + 0.999f*tr.t*dz);
			nContact =	tr.a1*v3d(renderInfo->n1[0], renderInfo->n1[1], renderInfo->n1[2]) +
						tr.a2*v3d(renderInfo->n2[0], renderInfo->n2[1], renderInfo->n2[2]) +
						tr.a3*v3d(renderInfo->n3[0], renderInfo->n3[1], renderInfo->n3[2]);

			// transform contact into world space (from model space)
			// pContact = scene->meshes[tr.mesh]->tree->matrix.Transform(pContact);
			// pContact = scene->meshes[tr.mesh]->matrix.Transform(pContact);
			nContact = scene->meshes[tr.mesh]->tree->matrix.NormalTransform(nContact);
			nContact = scene->meshes[tr.mesh]->matrix.NormalTransform(nContact);
			nContact = vmNormalize(nContact);

			// check to see if we hit the backside of the physical plane (not interpolated normal)
			if (vmDotProduct(nd, nPlane)<0) {
				nContact = -nContact;
				nPlane = -nPlane;
				backSide = true;
			}

			//  pContact = pContact + 0.0001f*nContact;	// to prevent poly shadowing itself due to rounding error

			// get surface color
			if (mat->tex) {
				tcContact = tr.a1*v3d(renderInfo->tc1[0], renderInfo->tc1[1], renderInfo->tc1[2]) +
							tr.a2*v3d(renderInfo->tc2[0], renderInfo->tc2[1], renderInfo->tc2[2]) +
							tr.a3*v3d(renderInfo->tc3[0], renderInfo->tc3[1], renderInfo->tc3[2]);
				eosSampleTexture(*(mat->tex), tcContact.x, tcContact.y, surfaceCol);
			} else {
				memcpy(surfaceCol, mat->color, 4*sizeof(float));
			}

			// if we need a reflection vector, calcualte it here
			// for both specular and reflection component
			if (mat->specular>0.0f || mat->glossReflection>0.0f || mat->metalicReflection>0.0f) {
				rVect = (2.0f * vmDotProduct(nd, nContact))*nContact - nd;
				rVect = vmNormalize(rVect);
				fresnelTerm = 1.0f - (nd*nContact);
			}

			// see if we'll need to handle ambient occlusion
			if (settings.flags & EOS_RENDERER_AMBIENT_OCCLUSION) {
				for (i=0;i<scene->nLights;i++) {
					if (scene->lights[i]->type == EOS_TYPE_AMBIENT || scene->lights[i]->type == EOS_TYPE_GRADIENT) {
						needsAO = true;
					}
				}
			}
			if (needsAO) {
				eosRendererGetAOValue(scene, pContact, nPlane, AOColor, settings);
			}

			// get diffuse & specular light
			diffuseCol[0] = 0.0f;	diffuseCol[1] = 0.0f;	diffuseCol[2] = 0.0f;	diffuseCol[3] = 1.0f;
			specularCol[0] = specularCol[1] = specularCol[2] = 0.0f;
			if (mat->unshaded) {
				diffuseCol[0] = diffuseCol[1] = diffuseCol[2] = diffuseCol[3] = 1.0f;
			} else {
				for (i=0;i<scene->nLights;i++) {
					eosLight * light = scene->lights[i];
					if (light->type == EOS_TYPE_AMBIENT) {
						if (needsAO) {
							diffuseCol[0] += AOColor[0]*light->col[0];
							diffuseCol[1] += AOColor[1]*light->col[1];
							diffuseCol[2] += AOColor[2]*light->col[2];
						} else {
							diffuseCol[0] += light->col[0];
							diffuseCol[1] += light->col[1];
							diffuseCol[2] += light->col[2];
						}
					} else if (light->type == EOS_TYPE_GRADIENT) {
						float dp = vmDotProduct(vmNormalize(v3d(light->d[0], light->d[1], light->d[2])), nContact);
						dp = 0.5f + 0.5f*dp;
						if (needsAO) {
							diffuseCol[0] += AOColor[0]*(light->col[0]*dp + light->col2[0]*(1.0f-dp));
							diffuseCol[1] += AOColor[1]*(light->col[1]*dp + light->col2[1]*(1.0f-dp));
							diffuseCol[2] += AOColor[2]*(light->col[2]*dp + light->col2[2]*(1.0f-dp));
						} else {
							diffuseCol[0] += light->col[0]*dp + light->col2[0]*(1.0f-dp);
							diffuseCol[1] += light->col[1]*dp + light->col2[1]*(1.0f-dp);
							diffuseCol[2] += light->col[2]*dp + light->col2[2]*(1.0f-dp);
						}
					} else if (light->type == EOS_TYPE_DIRECTIONAL) {
						SVect3d lVect = -v3d(light->d[0], light->d[1], light->d[2]);
						float dp = vmDotProduct(vmNormalize(lVect), nContact);
						if (dp>0) {
							float castShadow[3];
							eosRendererGetShadowColor(scene, pContact.x, pContact.y, pContact.z, lVect.x, lVect.y, lVect.z, -1.0f, castShadow);
							diffuseCol[0] += dp*light->col[0]*castShadow[0];
							diffuseCol[1] += dp*light->col[1]*castShadow[1];
							diffuseCol[2] += dp*light->col[2]*castShadow[2];
							if (mat->specular>0.0f) {
								float intensity = vmDotProduct(-vmNormalize(lVect), -rVect);
								if (intensity > 0.0f) {
									intensity = powf(intensity, 50.0f);
									intensity *= mat->specular;
									specularCol[0] += light->col[0]*intensity*castShadow[0];
									specularCol[1] += light->col[1]*intensity*castShadow[1];
									specularCol[2] += light->col[2]*intensity*castShadow[2];
								}
							}
						}
					} else if (light->type == EOS_TYPE_DIRECTIONAL_SOFT) {
						SVect3d lVect = -v3d(light->d[0], light->d[1], light->d[2]);
						float dp = vmDotProduct(vmNormalize(lVect), nContact);
						if (dp>0) {
							float castShadow[3];
							eosRendererGetSoftShadowColor(scene, pContact, *light, castShadow, settings);
							diffuseCol[0] += dp*light->col[0]*castShadow[0];
							diffuseCol[1] += dp*light->col[1]*castShadow[1];
							diffuseCol[2] += dp*light->col[2]*castShadow[2];
							if (mat->specular>0.0f) {
								float intensity = vmDotProduct(-vmNormalize(lVect), -rVect);
								if (intensity > 0.0f) {
									intensity = powf(intensity, 50.0f);
									intensity *= mat->specular;
									specularCol[0] += light->col[0]*intensity*castShadow[0];
									specularCol[1] += light->col[1]*intensity*castShadow[1];
									specularCol[2] += light->col[2]*intensity*castShadow[2];
								}
							}
						}
					} else if (light->type == EOS_TYPE_POINT || light->type == EOS_TYPE_SPOTLIGHT) {
						bool inRange = true;
						float falloffPct = 1.0f;
						SVect3d lVect = v3d(light->p[0], light->p[1], light->p[2]) - pContact;

						if (light->type == EOS_TYPE_SPOTLIGHT) {
							float spotDP = vmDotProduct(vmNormalize(lVect), -vmNormalize(v3d(light->d[0], light->d[1], light->d[2])));
							if (spotDP<light->cosFov) inRange = false;
						}
						if (inRange && light->r > 0.0f) {
							float r2 = vmDotProduct(lVect, lVect);
							if (r2 == 0.0f) {
								// avoid division by zero
								falloffPct = 1000000.0f;
							} else if (!eosPointInAABB(light->box, pContact.x, pContact.y, pContact.z)) {
								inRange = (r2<light->r2)?true:false;
							} else if (inRange) {
								if (light->falloff == EOS_FALLOFF_LINEAR) {
									float d = (float)sqrt(r2);
									falloffPct = 1.0f - d/light->r;
								} else if (light->falloff == EOS_FALLOFF_INVERSE) {
									float d = (float)sqrt(r2);
									falloffPct = 1.0f/d - 1.0f/light->r;
								} else if (light->falloff == EOS_FALLOFF_INVERSE_SQUARED) {
									falloffPct = 1.0f/r2 - 1.0f/light->r2;
								} else if (light->falloff == EOS_FALLOFF_ORB_LIGHT) {
									falloffPct = light->fov / (r2 + light->fov);
								}
							}
							if (falloffPct < 0.0f) falloffPct = 0.0f;
						} else if (light->falloff == EOS_FALLOFF_ORB_LIGHT) {
							float r2 = vmDotProduct(lVect, lVect);
							falloffPct = light->fov / (r2 + light->fov);
						}
						float dp = vmDotProduct(vmNormalize(lVect), nContact);
						if (inRange && dp>0.0f) {
							float castShadow[3];
							eosRendererGetShadowColor(scene, pContact.x, pContact.y, pContact.z, lVect.x, lVect.y, lVect.z, 1.0f, castShadow);
							diffuseCol[0] += dp*falloffPct*light->col[0]*castShadow[0];
							diffuseCol[1] += dp*falloffPct*light->col[1]*castShadow[1];
							diffuseCol[2] += dp*falloffPct*light->col[2]*castShadow[2];
							if (mat->specular>0.0f) {
								float intensity = vmDotProduct(-vmNormalize(lVect), -rVect);
								if (intensity > 0.0f) {
									intensity = powf(intensity, 50.0f);
									intensity *= mat->specular;
									specularCol[0] += falloffPct*light->col[0]*intensity*castShadow[0];
									specularCol[1] += falloffPct*light->col[1]*intensity*castShadow[1];
									specularCol[2] += falloffPct*light->col[2]*intensity*castShadow[2];
								}
							}
						}
					}
				}
			}

			// if the alpha is not 100%, we need the color behind this point
			if (surfaceCol[3]<1.0f) {
				SVect3d tempP = v3d(dx, dy, dz);
				tempP = v3d(px, py, pz) + tempP*tr.t + 0.001f*vmNormalize(tempP);
				eosRendererTraceColor(scene, tempP.x, tempP.y, tempP.z, dx, dy, dz, colorBehind,level, settings);
			}

			// if we need to get a reflection, get it
			if (mat->glossReflection>0.0f || mat->metalicReflection>0.0f) {
				eosRendererTraceColor(scene, pContact.x, pContact.y, pContact.z, rVect.x, rVect.y, rVect.z, colorReflect, level-1, settings);
				fresnelTerm = (fresnelTerm*mat->fresnelTerm + 1.0f - mat->fresnelTerm);
				fresnelTerm = powf(fresnelTerm, 3.0f);
			}

			// calculate fog
			if (settings.fogRate>0.0f) {
				fogPct = 1.0f - powf(1.0f-settings.fogRate, vmLength(v3d(dx, dy, dz)*tr.t));
			}

			// blend all our colors in accordance with the surface
			eosRendererBlendColors(mat, surfaceCol, diffuseCol, specularCol, colorBehind, colorReflect, fresnelTerm, fogPct, settings.fogColor, returnColor);


		} else {
			if (scene->env) {
				float tempCol[4];
				eosSampleEnvironmentMap(scene->env, dx, dy, dz, tempCol);
				returnColor[0] = tempCol[0];
				returnColor[1] = tempCol[1];
				returnColor[2] = tempCol[2];
			} else {
				returnColor[0] = returnColor[1] = returnColor[2] = 0.0f;
			}
		}
	}
}

/// <summary>Takes 4 colors (current pixel and neighbors), decides if antialiasing is needed, and if so
/// subdivides pixel into 4 quadrants and recurses</summary>
/// <returns>(void)</returns>
void eosStandardAntialias(eosScene *scene, float *cp, float *cd1, float *cd2, float *cd3, float *cd4,
	float *col1, float *col2, float *col3, float *col4, float *outputCol, int level, const eosRendererSettings &settings) {

	if (level == 0 ) {
		memcpy(outputCol, col1, 3*sizeof(float));
	} else {
		float diff1, diff2, diff3, avg;
		avg = (col1[0] + col1[1] + col1[2])*0.33f;
		diff1 = (fabsf(col1[0]-col2[0]) + fabsf(col1[1]-col2[1]) + fabsf(col1[2]-col2[2]))*0.33f;
		diff2 = (fabsf(col1[0]-col3[0]) + fabsf(col1[1]-col3[1]) + fabsf(col1[2]-col3[2]))*0.33f;
		diff3 = (fabsf(col1[0]-col4[0]) + fabsf(col1[1]-col4[1]) + fabsf(col1[2]-col4[2]))*0.33f;

		diff1 = diff1/avg;
		diff2 = diff2/avg;
		diff3 = diff3/avg;
	
		if ((diff1>settings.antialiasThreshold) ||
			(diff2>settings.antialiasThreshold) ||
			(diff3>settings.antialiasThreshold)) {
			float cdTop[3];
			float cdLeft[3];
			float cdCenter[3];
			float cdRight[3];
			float cdBottom[3];

			float colTop[3];
			float colLeft[3];
			float colCenter[3];
			float colRight[3];
			float colBottom[3];

			cdTop[0] = 0.5f*(cd1[0]+cd2[0]);
			cdTop[1] = 0.5f*(cd1[1]+cd2[1]);
			cdTop[2] = 0.5f*(cd1[2]+cd2[2]);
			eosRendererTraceColor(scene, cp[0], cp[1], cp[2], cdTop[0], cdTop[1], cdTop[2], colTop, settings.levels, settings);

			cdLeft[0] = 0.5f*(cd1[0]+cd3[0]);
			cdLeft[1] = 0.5f*(cd1[1]+cd3[1]);
			cdLeft[2] = 0.5f*(cd1[2]+cd3[2]);
			eosRendererTraceColor(scene, cp[0], cp[1], cp[2], cdLeft[0], cdLeft[1], cdLeft[2], colLeft, settings.levels, settings);

			cdCenter[0] = 0.5f*(cd1[0]+cd4[0]);
			cdCenter[1] = 0.5f*(cd1[1]+cd4[1]);
			cdCenter[2] = 0.5f*(cd1[2]+cd4[2]);
			eosRendererTraceColor(scene, cp[0], cp[1], cp[2], cdCenter[0], cdCenter[1], cdCenter[2], colCenter, settings.levels, settings);

			cdRight[0] = 0.5f*(cd2[0]+cd4[0]);
			cdRight[1] = 0.5f*(cd2[1]+cd4[1]);
			cdRight[2] = 0.5f*(cd2[2]+cd4[2]);
			eosRendererTraceColor(scene, cp[0], cp[1], cp[2], cdRight[0], cdRight[1], cdRight[2], colRight, settings.levels, settings);

			cdBottom[0] = 0.5f*(cd3[0]+cd4[0]);
			cdBottom[1] = 0.5f*(cd3[1]+cd4[1]);
			cdBottom[2] = 0.5f*(cd3[2]+cd4[2]);
			eosRendererTraceColor(scene, cp[0], cp[1], cp[2], cdBottom[0], cdBottom[1], cdBottom[2], colBottom, settings.levels, settings);

			float colQ1[3], colQ2[3], colQ3[3], colQ4[3];
			eosStandardAntialias(scene, cp, cd1, cdTop, cdLeft, cdCenter, col1, colTop, colLeft, colCenter, colQ1, level-1, settings);
			eosStandardAntialias(scene, cp, cdTop, cd2, cdCenter, cdRight, colTop, col2, colCenter, colRight, colQ2, level-1, settings);
			eosStandardAntialias(scene, cp, cdLeft, cdCenter, cd3, cdBottom, colLeft, colCenter, col3, colBottom, colQ3, level-1, settings);
			eosStandardAntialias(scene, cp, cdCenter, cdRight, cdBottom, cd4, colCenter, colRight, colBottom, col4, colQ4, level-1, settings);

			outputCol[0] = 0.25f*(colQ1[0] + colQ2[0] + colQ3[0] + colQ4[0]);
			outputCol[1] = 0.25f*(colQ1[1] + colQ2[1] + colQ3[1] + colQ4[1]);
			outputCol[2] = 0.25f*(colQ1[2] + colQ2[2] + colQ3[2] + colQ4[2]);

		} else {
			memcpy(outputCol, col1, 3*sizeof(float));
		}
	}
}

/// <summary>Renders a scene using standard methods, single threaded, and puts results in outputBuffer.</summary>
/// <returns>(void)</returns>
void eosRenderScene_STANDARD_SINGLETHREADED (eosScene *scene, eosCamera *camera, float *outputBuffer, const eosRendererSettings &settings) {
	unsigned int		i, j;			// loop counters
	float				*buff;			// pointer to current pixel
	float				cp[3];			// camera position
	float				cd[3];			// camera direction (start getting 4 corners and interpolating later)

	buff = outputBuffer;
	for (j=0;j<camera->h;j++) {
		for (i=0;i<camera->w;i++) {
			eosCameraProject(*camera, (float)i, (float)j, cp, cd);
			eosRendererTraceColor(scene, cp[0], cp[1], cp[2], cd[0], cd[1], cd[2], buff, settings.levels, settings);
			buff += 3;
		}
	}

	buff = outputBuffer;
	if (settings.flags & EOS_RENDERER_ANTIALIAS) {
		for (j=0;j<camera->h-1;j++) {
			buff = 3*camera->w*j + outputBuffer;
			for (i=0;i<camera->w-1;i++) {
				float outputCol[3];
				float cd1[3], cd2[3], cd3[3], cd4[3];
				eosCameraProject(*camera, (float)(i+0), (float)(j+0), cp, cd1);
				eosCameraProject(*camera, (float)(i+1), (float)(j+0), cp, cd2);
				eosCameraProject(*camera, (float)(i+0), (float)(j+1), cp, cd3);
				eosCameraProject(*camera, (float)(i+1), (float)(j+1), cp, cd4);
				eosStandardAntialias(scene, cp, cd1, cd2, cd3, cd4, buff, buff+3, buff+3*camera->w, buff+3*camera->w+3, outputCol, settings.antialiasLevels, settings);
				memcpy(buff, outputCol, 3*sizeof(float));
				buff+=3;
			}
		}
	}
}

/*
	+------------------+--------------------+
	| SINGLETHREADED   | SUBSAMPLE          |
	+------------------+--------------------+
*/


struct eosInterpolationInformation {
	bool				traced;				// whether or not this pixel's been traced
	eosScene			*scene;				// the scene we're tracing
	eosTraceResult		tr;					// the trace result at this pixel
	float				texCoord[2];		// texture coordinates
	float				surfaceCol[4];		// surface color
	float				diffuseCol[4];		// diffuse light color (TODO: make 3 floats - lots of places to change below)
	float				specularCol[3];		// specular color
	float				reflectionCol[3];	// reflection color
	float				colorBehind[3];		// color behind this contact
	float				fresnelTerm;		// fresnel term
	SVect3d				p, d;				// the ray p+dt that was traced
	float				fogPct;				// fog contribution
};

/// <summary>Traces a single ray and populates the interpolation information structure</summary>
/// <returns>(void)</returns>
void eosInterpolationGetSample(eosScene *scene, float *cp, float *cd, eosInterpolationInformation *out, const eosRendererSettings &settings) {

	// trace this ray, and mark it traced
	out->tr = eosTraceScene(scene, cp[0], cp[1], cp[2], cd[0], cd[1], cd[2], -1.0f);
	out->traced = true;
	out->p = v3d(cp[0], cp[1], cp[2]);
	out->d = v3d(cd[0], cd[1], cd[2]);
	out->scene = scene;

	// if we hit something, get the values we require
	if (out->tr.t>0.0f) {
			int						i;					// loop counters
			SVect3d					pContact;			// point of contact
			SVect3d					nContact;			// normal at contact
			SVect3d					nPlane;				// normal of the plane itself
			SVect3d					tcContact;			// texcoords at contact
			SVect3d					nd;					// negative trace direction (used more frequently than positive)
			SVect3d					rVect;				// reflection vector
			float					fresnelTerm;		// fresnel term
			float					surfaceCol[4];		// surface color
			eosMaterial *			mat;				// pointer to material
			eosTriangle *			tri;				// pointer to triangle
			eosTriangleRenderInfo *	renderInfo;			// pointer to render info
			float					diffuseCol[4];		// diffuse light color
			float					specularCol[3];		// specular light color
			float					colorBehind[3];		// color behind this contact (if alpha<1.0)
			float					colorReflect[3];	// color from reflection
			bool					backSide = false;
			bool					needsAO = false;
			float					AOColor[3];			// ambient occlusion multiplier

			nd = -vmNormalize(v3d(cd[0], cd[1], cd[2]));

			// get pointers to our triangle, render info, and material
			tri = &(scene->meshes[out->tr.mesh]->tris[out->tr.tri]);
			renderInfo = &(scene->meshes[out->tr.mesh]->renderInfo[out->tr.tri]);
			mat = renderInfo->mat;

			// get the plane normal in world space
			nPlane = v3d(tri->pnx, tri->pny, tri->pnz);
			nPlane = scene->meshes[out->tr.mesh]->tree->matrix.NormalTransform(nPlane);
			nPlane = scene->meshes[out->tr.mesh]->matrix.NormalTransform(nPlane);

			// use trace result to calcuate contact point and normal at contact
			pContact = v3d(cp[0] + 0.999f*out->tr.t*cd[0], cp[1] + 0.999f*out->tr.t*cd[1], cp[2] + 0.999f*out->tr.t*cd[2]);
			nContact =	out->tr.a1*v3d(renderInfo->n1[0], renderInfo->n1[1], renderInfo->n1[2]) +
						out->tr.a2*v3d(renderInfo->n2[0], renderInfo->n2[1], renderInfo->n2[2]) +
						out->tr.a3*v3d(renderInfo->n3[0], renderInfo->n3[1], renderInfo->n3[2]);

			// transform contact into world space (from model space)
			nContact = scene->meshes[out->tr.mesh]->tree->matrix.NormalTransform(nContact);
			nContact = scene->meshes[out->tr.mesh]->matrix.NormalTransform(nContact);
			nContact = vmNormalize(nContact);

			// check to see if we hit the backside of the physical plane (not interpolated normal)
			if (vmDotProduct(nd, nPlane)<0) {
				nContact = -nContact;
				nPlane = -nPlane;
				backSide = true;
			}

			// get surface color
			if (mat->tex) {
				tcContact = out->tr.a1*v3d(renderInfo->tc1[0], renderInfo->tc1[1], renderInfo->tc1[2]) +
							out->tr.a2*v3d(renderInfo->tc2[0], renderInfo->tc2[1], renderInfo->tc2[2]) +
							out->tr.a3*v3d(renderInfo->tc3[0], renderInfo->tc3[1], renderInfo->tc3[2]);
				eosSampleTexture(*(mat->tex), tcContact.x, tcContact.y, surfaceCol);
				// save texCoord in interpolation info
				out->texCoord[0] = tcContact.x;
				out->texCoord[1] = tcContact.y;
			} else {
				memcpy(surfaceCol, mat->color, 4*sizeof(float));
			}
			// save surfaceCol in interpolation info
			memcpy(out->surfaceCol, surfaceCol, 4*sizeof(float));

			// if we need a reflection vector, calcualte it here
			// for both specular and reflection component
			if (mat->specular>0.0f || mat->glossReflection>0.0f || mat->metalicReflection>0.0f) {
				rVect = (2.0f * vmDotProduct(nd, nContact))*nContact - nd;
				rVect = vmNormalize(rVect);
				fresnelTerm = 1.0f - (nd*nContact);
			}

			// see if we'll need to handle ambient occlusion
			if (settings.flags & EOS_RENDERER_AMBIENT_OCCLUSION) {
				for (i=0;i<scene->nLights;i++) {
					if (scene->lights[i]->type == EOS_TYPE_AMBIENT || scene->lights[i]->type == EOS_TYPE_GRADIENT) {
						needsAO = true;
					}
				}
			}
			if (needsAO) {
				eosRendererGetAOValue(scene, pContact, nPlane, AOColor, settings);
			}

			// get diffuse & specular light
			diffuseCol[0] = 0.0f;	diffuseCol[1] = 0.0f;	diffuseCol[2] = 0.0f;	diffuseCol[3] = 1.0f;
			specularCol[0] = specularCol[1] = specularCol[2] = 0.0f;
			if (mat->unshaded) {
				diffuseCol[0] = diffuseCol[1] = diffuseCol[2] = diffuseCol[3] = 1.0f;
			} else {
				for (i=0;i<scene->nLights;i++) {
					eosLight * light = scene->lights[i];
					if (light->type == EOS_TYPE_AMBIENT) {
						if (needsAO) {
							diffuseCol[0] += AOColor[0]*light->col[0];
							diffuseCol[1] += AOColor[1]*light->col[1];
							diffuseCol[2] += AOColor[2]*light->col[2];
						} else {
							diffuseCol[0] += light->col[0];
							diffuseCol[1] += light->col[1];
							diffuseCol[2] += light->col[2];
						}
					} else if (light->type == EOS_TYPE_GRADIENT) {
						float dp = vmDotProduct(vmNormalize(v3d(light->d[0], light->d[1], light->d[2])), nContact);
						dp = 0.5f + 0.5f*dp;
						if (needsAO) {
							diffuseCol[0] += AOColor[0]*(light->col[0]*dp + light->col2[0]*(1.0f-dp));
							diffuseCol[1] += AOColor[1]*(light->col[1]*dp + light->col2[1]*(1.0f-dp));
							diffuseCol[2] += AOColor[2]*(light->col[2]*dp + light->col2[2]*(1.0f-dp));
						} else {
							diffuseCol[0] += light->col[0]*dp + light->col2[0]*(1.0f-dp);
							diffuseCol[1] += light->col[1]*dp + light->col2[1]*(1.0f-dp);
							diffuseCol[2] += light->col[2]*dp + light->col2[2]*(1.0f-dp);
						}
					} else if (light->type == EOS_TYPE_DIRECTIONAL) {
						SVect3d lVect = -v3d(light->d[0], light->d[1], light->d[2]);
						float dp = vmDotProduct(vmNormalize(lVect), nContact);
						if (dp>0) {
							float castShadow[3];
							eosRendererGetShadowColor(scene, pContact.x, pContact.y, pContact.z, lVect.x, lVect.y, lVect.z, -1.0f, castShadow);
							diffuseCol[0] += dp*light->col[0]*castShadow[0];
							diffuseCol[1] += dp*light->col[1]*castShadow[1];
							diffuseCol[2] += dp*light->col[2]*castShadow[2];
							if (mat->specular>0.0f) {
								float intensity = vmDotProduct(-vmNormalize(lVect), -rVect);
								if (intensity > 0.0f) {
									intensity = powf(intensity, 50.0f);
									intensity *= mat->specular;
									specularCol[0] += light->col[0]*intensity*castShadow[0];
									specularCol[1] += light->col[1]*intensity*castShadow[1];
									specularCol[2] += light->col[2]*intensity*castShadow[2];
								}
							}
						}
					} else if (light->type == EOS_TYPE_DIRECTIONAL_SOFT) {
						SVect3d lVect = -v3d(light->d[0], light->d[1], light->d[2]);
						float dp = vmDotProduct(vmNormalize(lVect), nContact);
						if (dp>0) {
							float castShadow[3];
							eosRendererGetSoftShadowColor(scene, pContact, *light, castShadow, settings);
							diffuseCol[0] += dp*light->col[0]*castShadow[0];
							diffuseCol[1] += dp*light->col[1]*castShadow[1];
							diffuseCol[2] += dp*light->col[2]*castShadow[2];
							if (mat->specular>0.0f) {
								float intensity = vmDotProduct(-vmNormalize(lVect), -rVect);
								if (intensity > 0.0f) {
									intensity = powf(intensity, 50.0f);
									intensity *= mat->specular;
									specularCol[0] += light->col[0]*intensity*castShadow[0];
									specularCol[1] += light->col[1]*intensity*castShadow[1];
									specularCol[2] += light->col[2]*intensity*castShadow[2];
								}
							}
						}
					} else if (light->type == EOS_TYPE_POINT || light->type == EOS_TYPE_SPOTLIGHT) {
						bool inRange = true;
						float falloffPct = 1.0f;
						SVect3d lVect = v3d(light->p[0], light->p[1], light->p[2]) - pContact;

						if (light->type == EOS_TYPE_SPOTLIGHT) {
							float spotDP = vmDotProduct(vmNormalize(lVect), -vmNormalize(v3d(light->d[0], light->d[1], light->d[2])));
							if (spotDP<light->cosFov) inRange = false;
						}
						if (inRange && light->r > 0.0f) {
							float r2 = vmDotProduct(lVect, lVect);
							if (r2 == 0.0f) {
								// avoid division by zero
								falloffPct = 1000000.0f;
							} else if (!eosPointInAABB(light->box, pContact.x, pContact.y, pContact.z)) {
								inRange = (r2<light->r2)?true:false;
							} else if (inRange) {
								if (light->falloff == EOS_FALLOFF_LINEAR) {
									float d = (float)sqrt(r2);
									falloffPct = 1.0f - d/light->r;
								} else if (light->falloff == EOS_FALLOFF_INVERSE) {
									float d = (float)sqrt(r2);
									falloffPct = 1.0f/d - 1.0f/light->r;
								} else if (light->falloff == EOS_FALLOFF_INVERSE_SQUARED) {
									falloffPct = 1.0f/r2 - 1.0f/light->r2;
								} else if (light->falloff == EOS_FALLOFF_ORB_LIGHT) {
									falloffPct = light->fov / (r2 + light->fov);
								}
							}
							if (falloffPct < 0.0f) falloffPct = 0.0f;
						} else if (light->falloff == EOS_FALLOFF_ORB_LIGHT) {
							float r2 = vmDotProduct(lVect, lVect);
							falloffPct = light->fov / (r2 + light->fov);
						}
						float dp = vmDotProduct(vmNormalize(lVect), nContact);
						if (inRange && dp>0.0f) {
							float castShadow[3];
							eosRendererGetShadowColor(scene, pContact.x, pContact.y, pContact.z, lVect.x, lVect.y, lVect.z, 1.0f, castShadow);
							diffuseCol[0] += dp*falloffPct*light->col[0]*castShadow[0];
							diffuseCol[1] += dp*falloffPct*light->col[1]*castShadow[1];
							diffuseCol[2] += dp*falloffPct*light->col[2]*castShadow[2];
							if (mat->specular>0.0f) {
								float intensity = vmDotProduct(-vmNormalize(lVect), -rVect);
								if (intensity > 0.0f) {
									intensity = powf(intensity, 50.0f);
									intensity *= mat->specular;
									specularCol[0] += falloffPct*light->col[0]*intensity*castShadow[0];
									specularCol[1] += falloffPct*light->col[1]*intensity*castShadow[1];
									specularCol[2] += falloffPct*light->col[2]*intensity*castShadow[2];
								}
							}
						}
					}
				}
			}
			// save diffuse and specular in interpolation info
			memcpy(out->diffuseCol, diffuseCol, 4*sizeof(float));
			memcpy(out->specularCol, specularCol, 3*sizeof(float));

			// if the alpha is not 100%, we need the color behind this point
			if (surfaceCol[3]<1.0f) {
				SVect3d tempP = v3d(cd[0], cd[1], cd[2]);
				tempP = v3d(cp[0], cp[1], cp[2]) + tempP*out->tr.t + 0.001f*vmNormalize(tempP);
				eosRendererTraceColor(scene, tempP.x, tempP.y, tempP.z, cd[0], cd[1], cd[2], colorBehind, settings.levels, settings);
				// save in interpolation info
				memcpy(out->colorBehind, colorBehind, 3*sizeof(float));
			}

			// if we need to get a reflection, get it
			if (mat->glossReflection>0.0f || mat->metalicReflection>0.0f) {
				eosRendererTraceColor(scene, pContact.x, pContact.y, pContact.z, rVect.x, rVect.y, rVect.z, colorReflect, settings.levels-1, settings);
				fresnelTerm = (fresnelTerm*mat->fresnelTerm + 1.0f - mat->fresnelTerm);
				fresnelTerm = powf(fresnelTerm, 3.0f);
				// save in interpolation info
				memcpy(out->reflectionCol, colorReflect, 3*sizeof(float));
				out->fresnelTerm = fresnelTerm;
			}

			// calculate fog
			if (settings.fogRate > 0.0f) {
				out->fogPct = 1.0f - powf(1.0f-settings.fogRate, vmLength(out->tr.t*out->d));
			}
	}
}

/// <summary>Tells the subdivision routines whether or not a block needs to be subdivided or if
/// we can go ahead and interpolate</summary>
/// <returns>true if subdivision is requied, false if OK to interpolate</returns>
bool eosSubdivisionNeeded(eosInterpolationInformation *c1, eosInterpolationInformation *c2, eosInterpolationInformation *c3, eosInterpolationInformation *c4, const eosRendererSettings &settings) {
	bool hit1, hit2, hit3, hit4;

	// first, make sure they all hit SOMETHING or all hit NOTHING

	hit1 = c1->tr.t>=0?true:false;
	hit2 = c2->tr.t>=0?true:false;
	hit3 = c3->tr.t>=0?true:false;
	hit4 = c4->tr.t>=0?true:false;

	if ((hit1!=hit2) || (hit1!=hit3) || (hit1!=hit4)) return true;	// some hit things, some didn't: divide further
	if (!hit1) return false; // all hit nothing, no further division needed

	// next, make sure they all the same thing

	if ((c1->tr.mesh != c2->tr.mesh) ||
		(c1->tr.mesh != c3->tr.mesh) ||
		(c1->tr.mesh != c4->tr.mesh)) {
			return true;
	}

	// next, make sure they all hit the same material
	
	eosTriangleRenderInfo *renderInfo1 = &(c1->scene->meshes[c1->tr.mesh]->renderInfo[c1->tr.tri]);
	eosTriangleRenderInfo *renderInfo2 = &(c2->scene->meshes[c2->tr.mesh]->renderInfo[c2->tr.tri]);
	eosTriangleRenderInfo *renderInfo3 = &(c3->scene->meshes[c3->tr.mesh]->renderInfo[c3->tr.tri]);
	eosTriangleRenderInfo *renderInfo4 = &(c4->scene->meshes[c4->tr.mesh]->renderInfo[c4->tr.tri]);

	if ((renderInfo1->mat != renderInfo2->mat) || (renderInfo1->mat!=renderInfo3->mat) || (renderInfo1->mat!=renderInfo4->mat)) return true;
	
	// next, make sure they're all in the same tracerGroup
	// (this can be used to create explicit seams, to mark texture seams,
	// or to set apart smoothing groups)
	if ((renderInfo1->tracerGroup != renderInfo2->tracerGroup) ||
		(renderInfo1->tracerGroup != renderInfo3->tracerGroup) ||
		(renderInfo1->tracerGroup != renderInfo4->tracerGroup)) {
			return true;
	}

	// make sure that they're all either translucent or not
	// (if they're all translucent we'll need to interpolate colorBehind,
	// if they're all not, we won't - but if there's a mix we'll be
	// interpolating some good values with some uninitialized memory)
	unsigned int alphaCheck = 0x00;
	alphaCheck |= c1->surfaceCol[3]<1.0f?0x01:0x10;
	alphaCheck = alphaCheck<<1;
	alphaCheck |= c2->surfaceCol[3]<1.0f?0x01:0x10;
	alphaCheck = alphaCheck<<1;
	alphaCheck |= c3->surfaceCol[3]<1.0f?0x01:0x10;
	alphaCheck = alphaCheck<<1;
	alphaCheck |= c4->surfaceCol[3]<1.0f?0x01:0x10;

	if ((alphaCheck != 0xF0) && (alphaCheck != 0x0F)) return true;

	// if we're checking diffuse do it here
	if (settings.flags & EOS_RENDERER_SUBSAMPLE_DIFFUSE) {
		float diff1 = (c1->diffuseCol[0]+c1->diffuseCol[1]+c1->diffuseCol[2])/3.0f;
		float diff2 = (c2->diffuseCol[0]+c2->diffuseCol[1]+c2->diffuseCol[2])/3.0f;
		if (abs(diff1-diff2)>settings.subdivisionThreshold) return true;
		float diff3 = (c3->diffuseCol[0]+c3->diffuseCol[1]+c3->diffuseCol[2])/3.0f;
		if (abs(diff1-diff3)>settings.subdivisionThreshold) return true;
		if (abs(diff2-diff3)>settings.subdivisionThreshold) return true;
		float diff4 = (c4->diffuseCol[0]+c4->diffuseCol[1]+c4->diffuseCol[2])/3.0f;
		if (abs(diff1-diff4)>settings.subdivisionThreshold) return true;
		if (abs(diff2-diff4)>settings.subdivisionThreshold) return true;
		if (abs(diff3-diff4)>settings.subdivisionThreshold) return true;
	}

	// if we're checking specular do it here
	if ((settings.flags & EOS_RENDERER_SUBSAMPLE_SPECULAR) && (renderInfo1->mat->specular>0)) {
		float diff1 = (c1->specularCol[0]+c1->specularCol[1]+c1->specularCol[2])/3.0f;
		float diff2 = (c2->specularCol[0]+c2->specularCol[1]+c2->specularCol[2])/3.0f;
		if (abs(diff1-diff2)>settings.subdivisionThreshold) return true;
		float diff3 = (c3->specularCol[0]+c3->specularCol[1]+c3->specularCol[2])/3.0f;
		if (abs(diff1-diff3)>settings.subdivisionThreshold) return true;
		if (abs(diff2-diff3)>settings.subdivisionThreshold) return true;
		float diff4 = (c4->specularCol[0]+c4->specularCol[1]+c4->specularCol[2])/3.0f;
		if (abs(diff1-diff4)>settings.subdivisionThreshold) return true;
		if (abs(diff2-diff4)>settings.subdivisionThreshold) return true;
		if (abs(diff3-diff4)>settings.subdivisionThreshold) return true;
	}

	// if we're checking reflection do it here
	float rScalar = max(renderInfo1->mat->glossReflection, renderInfo1->mat->metalicReflection);
	if ((settings.flags & EOS_RENDERER_SUBSAMPLE_REFLECTION) && (rScalar>0)) {
		float diff1 = rScalar*(c1->reflectionCol[0]+c1->reflectionCol[1]+c1->reflectionCol[2])/3.0f;
		float diff2 = rScalar*(c2->reflectionCol[0]+c2->reflectionCol[1]+c2->reflectionCol[2])/3.0f;
		if (abs(diff1-diff2)>settings.reflectionThreshold) return true;
		float diff3 = rScalar*(c3->reflectionCol[0]+c3->reflectionCol[1]+c3->reflectionCol[2])/3.0f;
		if (abs(diff1-diff3)>settings.reflectionThreshold) return true;
		if (abs(diff2-diff3)>settings.reflectionThreshold) return true;
		float diff4 = rScalar*(c4->reflectionCol[0]+c4->reflectionCol[1]+c4->reflectionCol[2])/3.0f;
		if (abs(diff1-diff4)>settings.reflectionThreshold) return true;
		if (abs(diff2-diff4)>settings.reflectionThreshold) return true;
		if (abs(diff3-diff4)>settings.reflectionThreshold) return true;
	}

	// if we didn't trigger any of the subdivision conditions, then it's
	// OK to interpolate
	return false;
}

/// <summary>Takes a block of pixels, finds out if subdivision is needed, and either subdivides
/// into 4 sub-blocks, or interpolates data onto the frameBuffer</summary>
/// <returns>(void)</returns>
void eosSubsampleBlock(eosScene *scene, int blockSize, eosInterpolationInformation *intBuff, int bufferWidth, float *frameBuffer, unsigned int fbW, const eosRendererSettings &settings) {
	int halfBlockSize = blockSize>>1;

	// as long as the block is bigger than 1x1 pixel we check for subdivision
	if (blockSize>1) {
		bool subdivisionNeeded = false;
		
		// get corners - these all should be traced before this function is called
		eosInterpolationInformation *corner1 = intBuff;
		eosInterpolationInformation *corner2 = corner1+blockSize;
		eosInterpolationInformation *corner3 = corner1+blockSize*bufferWidth;
		eosInterpolationInformation *corner4 = corner2+blockSize*bufferWidth;

		// see if there's a need to subdivide
		subdivisionNeeded = eosSubdivisionNeeded(corner1, corner2, corner3, corner4, settings);

		if (subdivisionNeeded) {
			float cp[3], cd[3];
			SVect3d p, d;

			// get interpolation information for what will be the
			// corners of our 4 sub-blocks.
			eosInterpolationInformation *top = corner1 + halfBlockSize;
			eosInterpolationInformation *bottom = corner3 + halfBlockSize;
			eosInterpolationInformation *left = corner1 + (halfBlockSize*bufferWidth);
			eosInterpolationInformation *center = left + halfBlockSize;
			eosInterpolationInformation *right = left + blockSize;

			// trace top
			p = (corner1->p+corner2->p)*0.5f;
			d = (corner1->d+corner2->d)*0.5f;
			cp[0] = p.x; cp[1] = p.y; cp[2] = p.z;
			cd[0] = d.x; cd[1] = d.y; cd[2] = d.z;
			if (!top->traced) eosInterpolationGetSample(scene, cp, cd, top, settings);

			// trace bottom
			p = (corner3->p+corner4->p)*0.5f;
			d = (corner3->d+corner4->d)*0.5f;
			cp[0] = p.x; cp[1] = p.y; cp[2] = p.z;
			cd[0] = d.x; cd[1] = d.y; cd[2] = d.z;
			if (!bottom->traced) eosInterpolationGetSample(scene, cp, cd, bottom, settings);

			// trace left
			p = (corner1->p+corner3->p)*0.5f;
			d = (corner1->d+corner3->d)*0.5f;
			cp[0] = p.x; cp[1] = p.y; cp[2] = p.z;
			cd[0] = d.x; cd[1] = d.y; cd[2] = d.z;
			if (!left->traced) eosInterpolationGetSample(scene, cp, cd, left, settings);

			// trace right
			p = (corner2->p+corner4->p)*0.5f;
			d = (corner2->d+corner4->d)*0.5f;
			cp[0] = p.x; cp[1] = p.y; cp[2] = p.z;
			cd[0] = d.x; cd[1] = d.y; cd[2] = d.z;
			if (!right->traced) eosInterpolationGetSample(scene, cp, cd, right, settings);

			// trace center
			p = (corner1->p+corner4->p)*0.5f;
			d = (corner1->d+corner4->d)*0.5f;
			cp[0] = p.x; cp[1] = p.y; cp[2] = p.z;
			cd[0] = d.x; cd[1] = d.y; cd[2] = d.z;
			if (!center->traced) eosInterpolationGetSample(scene, cp, cd, center, settings);

			// subdivide
			eosSubsampleBlock(scene, halfBlockSize, corner1, bufferWidth, frameBuffer+3*(0), fbW, settings);
			eosSubsampleBlock(scene, halfBlockSize, top, bufferWidth, frameBuffer+3*(halfBlockSize), fbW, settings);
			eosSubsampleBlock(scene, halfBlockSize, left, bufferWidth, frameBuffer+3*(fbW*halfBlockSize), fbW, settings);
			eosSubsampleBlock(scene, halfBlockSize, center, bufferWidth, frameBuffer+3*(fbW*halfBlockSize + halfBlockSize), fbW, settings);

		// OK to interpolate, but we didn't hit anything
		} else if (intBuff->tr.t<0) {
			// interpolate environment map
			if (scene->env) {
				int i, j,k;
				SVect3d d, dLeft, dRight, dInc;
				float tmpColor[4];
				if (settings.flags&EOS_RENDERER_SUBSAMPLE_DEBUG) {
					k = 1;
				} else {
					k = blockSize;
				}
				for (j=0;j<k;j++) {
					dLeft = (corner1->d*(float)(blockSize-j) + corner3->d*(float)j)/(float)blockSize;
					dRight= (corner2->d*(float)(blockSize-j) + corner4->d*(float)j)/(float)blockSize;
					dInc = (dRight - dLeft)/(float)blockSize;
					d = dLeft;
					for (i=0;i<k;i++) {
						eosSampleEnvironmentMap(scene->env, d.x, d.y, d.z, tmpColor);
						memcpy(frameBuffer + 3*(j*fbW+i), tmpColor, 3*sizeof(float));
						d = d + dInc;
					}
				}
			} else {
				frameBuffer[0] = frameBuffer[1] = frameBuffer[2] = 0.0f;
			}

		// OK to interpolate
		} else {
			int i, j, k;
			float pctj, pcti;

			// these will hold the interpolated values for each
			// thing, and then passed to the blending function
			float surfaceCol[4];
			float diffuseCol[4];
			float specularCol[3];
			float colorBehind[3];
			float colorReflect[3];
			float fresnelTerm;
			float texCoord[2];
			float fogPct;

			// get pointer to material to keep our code from looking insane.
			eosMaterial *mat = scene->meshes[corner1->tr.mesh]->renderInfo[corner1->tr.tri].mat;
			// get the surface col - if there's a texture this'll get overwritten, but better than copying every time
			memcpy(surfaceCol, corner1->surfaceCol, 4*sizeof(float));

			if (settings.flags&EOS_RENDERER_SUBSAMPLE_DEBUG) {
				k = 1;
			} else {
				k = blockSize;
			}
			// iterate through block of pixels
			for (j=0;j<k;j++) {
				pctj = j/(float)blockSize;
				for (i=0;i<k;i++) {
					pcti = i/(float)blockSize;

					// surface col
					if (mat->tex) {
						texCoord[0] = (corner1->texCoord[0]*(1.0f-pcti) + corner2->texCoord[0]*pcti)*(1.0f-pctj) + (corner3->texCoord[0]*(1.0f-pcti) + corner4->texCoord[0]*pcti)*(pctj);
						texCoord[1] = (corner1->texCoord[1]*(1.0f-pcti) + corner2->texCoord[1]*pcti)*(1.0f-pctj) + (corner3->texCoord[1]*(1.0f-pcti) + corner4->texCoord[1]*pcti)*(pctj);
						eosSampleTexture(*(mat->tex), texCoord[0], texCoord[1], surfaceCol);
					}
					// diffuse col
					diffuseCol[0] = (corner1->diffuseCol[0]*(1.0f-pcti) + corner2->diffuseCol[0]*pcti)*(1.0f-pctj) + (corner3->diffuseCol[0]*(1.0f-pcti) + corner4->diffuseCol[0]*pcti)*(pctj);
					diffuseCol[1] = (corner1->diffuseCol[1]*(1.0f-pcti) + corner2->diffuseCol[1]*pcti)*(1.0f-pctj) + (corner3->diffuseCol[1]*(1.0f-pcti) + corner4->diffuseCol[1]*pcti)*(pctj);
					diffuseCol[2] = (corner1->diffuseCol[2]*(1.0f-pcti) + corner2->diffuseCol[2]*pcti)*(1.0f-pctj) + (corner3->diffuseCol[2]*(1.0f-pcti) + corner4->diffuseCol[2]*pcti)*(pctj);
					// specular col
					if (mat->specular > 0.0f) {
						specularCol[0] = (corner1->specularCol[0]*(1.0f-pcti) + corner2->specularCol[0]*pcti)*(1.0f-pctj) + (corner3->specularCol[0]*(1.0f-pcti) + corner4->specularCol[0]*pcti)*(pctj);
						specularCol[1] = (corner1->specularCol[1]*(1.0f-pcti) + corner2->specularCol[1]*pcti)*(1.0f-pctj) + (corner3->specularCol[1]*(1.0f-pcti) + corner4->specularCol[1]*pcti)*(pctj);
						specularCol[2] = (corner1->specularCol[2]*(1.0f-pcti) + corner2->specularCol[2]*pcti)*(1.0f-pctj) + (corner3->specularCol[2]*(1.0f-pcti) + corner4->specularCol[2]*pcti)*(pctj);
					}
					// reflection col
					if ((mat->glossReflection > 0.0f) || (mat->metalicReflection > 0.0f)) {
						colorReflect[0] = (corner1->reflectionCol[0]*(1.0f-pcti) + corner2->reflectionCol[0]*pcti)*(1.0f-pctj) + (corner3->reflectionCol[0]*(1.0f-pcti) + corner4->reflectionCol[0]*pcti)*(pctj);
						colorReflect[1] = (corner1->reflectionCol[1]*(1.0f-pcti) + corner2->reflectionCol[1]*pcti)*(1.0f-pctj) + (corner3->reflectionCol[1]*(1.0f-pcti) + corner4->reflectionCol[1]*pcti)*(pctj);
						colorReflect[2] = (corner1->reflectionCol[2]*(1.0f-pcti) + corner2->reflectionCol[2]*pcti)*(1.0f-pctj) + (corner3->reflectionCol[2]*(1.0f-pcti) + corner4->reflectionCol[2]*pcti)*(pctj);
						// fresnel term
						fresnelTerm = (corner1->fresnelTerm*(1.0f-pcti) + corner2->fresnelTerm*pcti)*(1.0f-pctj) + (corner3->fresnelTerm*(1.0f-pcti) + corner4->fresnelTerm*pcti)*(pctj);
					}
					// color behind
					if (corner1->surfaceCol[3]<1.0f) {
						colorBehind[0] = (corner1->colorBehind[0]*(1.0f-pcti) + corner2->colorBehind[0]*pcti)*(1.0f-pctj) + (corner3->colorBehind[0]*(1.0f-pcti) + corner4->colorBehind[0]*pcti)*(pctj);
						colorBehind[1] = (corner1->colorBehind[1]*(1.0f-pcti) + corner2->colorBehind[1]*pcti)*(1.0f-pctj) + (corner3->colorBehind[1]*(1.0f-pcti) + corner4->colorBehind[1]*pcti)*(pctj);
						colorBehind[2] = (corner1->colorBehind[2]*(1.0f-pcti) + corner2->colorBehind[2]*pcti)*(1.0f-pctj) + (corner3->colorBehind[2]*(1.0f-pcti) + corner4->colorBehind[2]*pcti)*(pctj);
					}
					// fog contribution
					fogPct = (corner1->fogPct*(1.0f-pcti) + corner2->fogPct*pcti)*(1.0f-pctj) + (corner3->fogPct*(1.0f-pcti) + corner4->fogPct*pcti)*pctj;

					eosRendererBlendColors(mat, surfaceCol, diffuseCol, specularCol, colorBehind, colorReflect, fresnelTerm, fogPct, settings.fogColor, frameBuffer + 3*(j*fbW+i));
				}
			}
		}

	} else {
		// make sure this 1 pixel is traced (in theory, this code should never execute)
		if (!intBuff->traced) {
			float cp[3];
			float cd[3];
			cp[0] = intBuff->p.x;
			cp[1] = intBuff->p.y;
			cp[2] = intBuff->p.z;
			cd[0] = intBuff->d.x;
			cd[1] = intBuff->d.y;
			cd[2] = intBuff->d.z;
			eosInterpolationGetSample(scene, cp, cd, intBuff, settings);
		}

		// if we hit something, blend the colors
		if (intBuff->tr.t>=0) {
			eosMaterial *mat = scene->meshes[intBuff->tr.mesh]->renderInfo[intBuff->tr.tri].mat;
			eosRendererBlendColors(mat, intBuff->surfaceCol, intBuff->diffuseCol, intBuff->specularCol, intBuff->colorBehind, intBuff->reflectionCol, intBuff->fresnelTerm, intBuff->fogPct, settings.fogColor, frameBuffer);
		// if not, sample the environment map
		} else if (scene->env) {
			float tempCol[4];
			eosSampleEnvironmentMap(scene->env, intBuff->d.x, intBuff->d.y, intBuff->d.z, tempCol);
			memcpy(frameBuffer, tempCol, 3*sizeof(float));
		}
	}
}

/// <summary>Renders a scene using subsampling</summary>
/// <returns>(void)</returns>
void eosRenderScene_SUBSAMPLE_SINGLETHREADED(eosScene *scene, eosCamera *camera, float *outputBuffer, const eosRendererSettings &settings) {
	unsigned int	i, j, innerJ;
	int				currentRow = 0;
	float			cp[3];
	float			cd[3];
	eosInterpolationInformation *intBuffer = (eosInterpolationInformation *)malloc((camera->w+1)*9*sizeof(eosInterpolationInformation));
	
	// first, get our initial values (but put them in the last row)
	for (i=0;i<=camera->w;i++) {
		if (i%8 == 0) {
			eosCameraProject(*camera, (float)i, 0.0f, cp, cd);
			eosInterpolationGetSample(scene, cp, cd, &(intBuffer[8*(camera->w+1)+i]), settings);
		} else {
			intBuffer[8*(camera->w+1)+i].traced = false;
		}
	}

	// now our interpolation loop
	for (j=0;j<camera->h;j+=8) {
		// clear rows 0 through 7
		for (innerJ=0;innerJ<8;innerJ++) {
			for (i=0;i<=camera->w;i++) {
				intBuffer[innerJ*(camera->w+1)+i].traced = false;
			}
		}
		// copy row 8 to row 0
		memcpy(intBuffer, intBuffer+(8*(camera->w+1)), (camera->w+1)*sizeof(eosInterpolationInformation));
		// trace row 8
		for (i=0;i<=camera->w;i++) {
			if (i%8 == 0) {
				eosCameraProject(*camera, (float)i, (float)(j+8), cp, cd);
				eosInterpolationGetSample(scene, cp, cd, &(intBuffer[8*(camera->w+1)+i]), settings);
			} else {
				intBuffer[8*(camera->w+1)+i].traced = false;
			}
		}

		// trace blocks
		for (i=0;i<camera->w;i+=8) {
			eosSubsampleBlock(scene, 8, intBuffer+i, camera->w+1, outputBuffer + (currentRow*camera->w+i)*3, camera->w, settings);
		}
		currentRow += 8;
	}
	free(intBuffer);
}

#ifdef WIN32

/*
	+------------------+--------------------+
	| MULTITHREADED    | STANDARD           |
	+------------------+--------------------+
*/

struct eosMultiThreadAntialiasInfo {
	eosScene *			scene;
	eosCamera *			camera;
	float *				outputBuffer;		// try shared buffer first
	int					threadNo;
	int					nThreads;
	eosRendererSettings	settings;
};

struct eosMultiThreadInfo {
	int					threadNo;			// which thread this one is
	int					nThreads;			// how many threads total
	eosScene			*scene;				// pointer to scene we're tracing
	eosCamera			*camera;			// pointer to camera view
	float				*outputBuffer;		// pointer to output colors to (each thread gets own buffer)
	eosRendererSettings settings;			// render settings
};

DWORD WINAPI eosRendererAntialiasThread(void *args) {
	eosMultiThreadAntialiasInfo *	info = (eosMultiThreadAntialiasInfo *)args;
	unsigned int					i, j;
	float *							buff;
	float							cp[3];
	float outputCol[3];
	float cd1[3], cd2[3], cd3[3], cd4[3];

	for (j=info->threadNo;j<info->camera->h-1;j+=info->nThreads) {
		buff = 3*info->camera->w*j + info->outputBuffer;
		for (i=0;i<info->camera->w-1;i++) {
			eosCameraProject(*(info->camera), (float)(i+0), (float)(j+0), cp, cd1);
			eosCameraProject(*(info->camera), (float)(i+1), (float)(j+0), cp, cd2);
			eosCameraProject(*(info->camera), (float)(i+0), (float)(j+1), cp, cd3);
			eosCameraProject(*(info->camera), (float)(i+1), (float)(j+1), cp, cd4);
			eosStandardAntialias(info->scene, cp, cd1, cd2, cd3, cd4, buff, buff+3, buff+3*info->camera->w, buff+3*info->camera->w+3, outputCol, info->settings.antialiasLevels, info->settings);
			memcpy(buff, outputCol, 3*sizeof(float));
			buff+=3;
		}
	}

	return 1;
}

/// <summary>Renders one thread's contribution to the final image. Threads draw
/// using scanline interleaving to distribute fill rate evenly</summary>
/// <returns>1</returns>
DWORD WINAPI eosRendererStandardThread(void *args) {
	eosMultiThreadInfo *info = (eosMultiThreadInfo *)args;
	unsigned int		i, j;			// loop counters
	float				*buff;			// pointer to current pixel
	float				cp[3];			// camera position
	float				cd[3];			// camera direction (start getting 4 corners and interpolating later)

	buff = info->outputBuffer;
	for (j=info->threadNo;j<info->camera->h;j+=info->nThreads) {
		for (i=0;i<info->camera->w;i++) {
			eosCameraProject(*(info->camera), (float)i, (float)j, cp, cd);
			eosRendererTraceColor(info->scene, cp[0], cp[1], cp[2], cd[0], cd[1], cd[2], buff, info->settings.levels, info->settings);
			buff+=3;
		}
	}

	return 1;
}

/// <summary>Allocates buffers, spawns threads, and collects results.</summary>
/// <returns>(void)</returns>
void eosRenderScene_STANDARD_MULTITHREADED (eosScene *scene, eosCamera *camera, float *outputBuffer, const eosRendererSettings &settings) {
	int i, j, threadLine;
	float *buff;

	eosMultiThreadInfo *	info;
	eosMultiThreadAntialiasInfo *	aInfo;
	HANDLE *				threadHandles;
	DWORD *					threadIDs;
	DWORD					WaitResult;

	info = (eosMultiThreadInfo *)malloc(settings.threads*sizeof(eosMultiThreadInfo));
	threadHandles = (HANDLE *)malloc(settings.threads*sizeof(HANDLE));
	threadIDs = (DWORD *)malloc(settings.threads*sizeof(DWORD));

	for (i=0;i<settings.threads;i++) {
		info[i].nThreads = settings.threads;
		info[i].threadNo = i;
		info[i].camera = camera;
		info[i].scene = scene;
		info[i].outputBuffer = (float *)malloc((1 + camera->w*camera->h*3/settings.threads)*sizeof(float));
		info[i].settings = settings;
	}

	
	for (i=0;i<settings.threads;i++) {
		threadHandles[i] = CreateThread(NULL, 0, eosRendererStandardThread, info+i, 0, threadIDs+i);
	}

	WaitResult = WaitForMultipleObjects(EOS_RENDERER_THREADS, threadHandles, TRUE, INFINITE);
	
	for (i=0;i<settings.threads;i++) {
		CloseHandle(threadHandles[i]);
	}


	buff = outputBuffer;
	threadLine=0;
	j=0;
	for (i=0;i<(int)camera->h;i++) {

		memcpy(buff, info[j].outputBuffer + threadLine*camera->w*3, camera->w*3*sizeof(float));

		buff += camera->w*3;
		j = (j+1)%EOS_RENDERER_THREADS;
		if (j==0) threadLine++;
	}

	for (i=0;i<settings.threads;i++) {
		free(info[i].outputBuffer);
	}

	if (settings.flags & EOS_RENDERER_ANTIALIAS) {
		aInfo = (eosMultiThreadAntialiasInfo *)malloc(settings.threads * sizeof(eosMultiThreadAntialiasInfo));

		for (i=0;i<settings.threads;i++) {
			aInfo[i].camera = camera;
			aInfo[i].nThreads = settings.threads;
			aInfo[i].outputBuffer = outputBuffer;
			aInfo[i].scene = scene;
			aInfo[i].threadNo = i;
			aInfo[i].settings = settings;
		}

		for (i=0;i<settings.threads;i++) {
			threadHandles[i] = CreateThread(NULL, 0, eosRendererAntialiasThread, aInfo+i, 0, threadIDs+i);
		}

		WaitResult = WaitForMultipleObjects(settings.threads, threadHandles, TRUE, INFINITE);

		for (i=0;i<settings.threads;i++) {
			CloseHandle(threadHandles[i]);
		}

		free(aInfo);
	}

	free(info);
	free(threadIDs);
	free(threadHandles);

}

/*
	+------------------+--------------------+
	| MULTITHREADED    | SUBSAMPLE          |
	+------------------+--------------------+
*/

struct eosSubsampleThreadInfo {
	SVect3d				p, d1, d2, d3, d4;		// ray information for 4 corners of output buffer
	float				*outputBuffer;			// output buffer (each thread gets own buffer)
	unsigned int		obW, obH;				// output buffer dimensions
	eosRendererSettings	settings;				// renderer settings
	eosScene			*scene;					// scene we're tracing
};

/// <summary>Uses subsampling to render the view described by args->p,d1,d2,d3,d4
/// into args->outputBuffer</summary>
/// <returns>1</returns>
DWORD WINAPI eosRendererSubsampleThread(void *args) {
	eosSubsampleThreadInfo *data = (eosSubsampleThreadInfo *)args;

	unsigned int	i, j, innerJ;
	SVect3d			d;
	float			cp[3], cd[3];
	float			pctX, pctY, ipctX, ipctY;
	eosInterpolationInformation *intBuffer = (eosInterpolationInformation *)malloc((data->obW+1)*9*sizeof(eosInterpolationInformation));
	
	// first, get our initial values (but put them in the last row)
	cp[0] = data->p.x;
	cp[1] = data->p.y;
	cp[2] = data->p.z;
	for (i=0;i<=data->obW;i++) {
		if (i%8 == 0) {
			pctX = (float)i/(float)data->obW;
			ipctX = 1.0f-pctX;
			d = data->d1*ipctX + data->d2*pctX;
			cd[0] = d.x; cd[1] = d.y; cd[2] = d.z;
			eosInterpolationGetSample(data->scene, cp, cd, &(intBuffer[8*(data->obW+1)+i]), data->settings);
		} else {
			intBuffer[8*(data->obW+1)+i].traced = false;
		}
	}

	// now our interpolation loop
	for (j=0;j<data->obH;j+=8) {
		pctY = (float)(j+8)/(float)data->obH;
		ipctY = 1.0f-pctY;
		// clear rows 0 through 7
		for (innerJ=0;innerJ<8;innerJ++) {
			for (i=0;i<=data->obW;i++) {
				intBuffer[innerJ*(data->obW+1)+i].traced = false;
			}
		}
		// copy row 8 to row 0
		memcpy(intBuffer, intBuffer+(8*(data->obW+1)), (data->obW+1)*sizeof(eosInterpolationInformation));
		// trace row 8
		for (i=0;i<=data->obW;i++) {
			if (i%8 == 0) {
				pctX = (float)i/(float)data->obW;
				ipctX = 1.0f-pctX;
				d = ipctY*(data->d1*ipctX + data->d2*pctX) + pctY*(data->d3*ipctX + data->d4*pctX);
				cd[0] = d.x; cd[1] = d.y; cd[2] = d.z;
				eosInterpolationGetSample(data->scene, cp, cd, &(intBuffer[8*(data->obW+1)+i]), data->settings);
			} else {
				intBuffer[8*(data->obW+1)+i].traced = false;
			}
		}

		// trace blocks
		for (i=0;i<data->obW;i+=8) {
			eosSubsampleBlock(data->scene, 8, intBuffer+i, data->obW+1, data->outputBuffer + (j*data->obW+i)*3, data->obW, data->settings);
		}
	}
	free(intBuffer);

	return 1;
}

/// <summary>Allocates storage, spawns threads, and collects results for subsample renderer. Cuts the outputBuffer into
/// vertical strips to minimize duplicated traces</summary>
/// <returns>(void)</returns>
void eosRenderScene_SUBSAMPLE_MULTITHREADED(eosScene *scene, eosCamera *camera, float *outputBuffer, const eosRendererSettings &settings) {
	int i, j;
	float cp[3], cd[3];
	int nThreads = settings.threads;
	int frameInc = camera->w/nThreads;
	float *buff;
	
	eosSubsampleThreadInfo *info = (eosSubsampleThreadInfo *)malloc(nThreads*sizeof(eosSubsampleThreadInfo));
	HANDLE *threadHandles = (HANDLE *)malloc(nThreads*sizeof(HANDLE));
	DWORD *threadIDs = (DWORD *)malloc(nThreads*sizeof(DWORD));

	// init thread arguments
	for (i=0;i<nThreads;i++) {
		// set the easy stuff
		info[i].scene = scene;
		info[i].settings = settings;
		info[i].obH = camera->h;
		info[i].obW = frameInc;
		// get direction vectors
		eosCameraProject(*camera, (float)(i*frameInc), 0.0f, cp, cd);
		info[i].p = v3d(cp[0], cp[1], cp[2]);
		info[i].d1 = v3d(cd[0], cd[1], cd[2]);
		eosCameraProject(*camera, (float)(i+1)*frameInc, 0.0f, cp, cd);
		info[i].d2 = v3d(cd[0], cd[1], cd[2]);
		eosCameraProject(*camera, (float)(i*frameInc), (float)camera->h, cp, cd);
		info[i].d3 = v3d(cd[0], cd[1], cd[2]);
		eosCameraProject(*camera, (float)(i+1)*frameInc, (float)camera->h, cp, cd);
		info[i].d4 = v3d(cd[0], cd[1], cd[2]);
		// allocate output buffer
		info[i].outputBuffer = (float *)malloc(3*frameInc*camera->h*sizeof(float));
		memset(info[i].outputBuffer, 0, 3*frameInc*camera->h*sizeof(float));
	}

	// spawn threads
	for (i=0;i<nThreads;i++) {
		threadHandles[i] = CreateThread(NULL, 0, eosRendererSubsampleThread, info+i, 0, threadIDs+i);
	}

	// wait for threads
	DWORD WaitResult = WaitForMultipleObjects(nThreads, threadHandles, TRUE, INFINITE);
	for (i=0;i<nThreads;i++) {
		CloseHandle(threadHandles[i]);
	}

	// copy frame data back to the outputBuffer
	buff = outputBuffer;
	memset(outputBuffer, 0, 3*camera->w*camera->h*sizeof(float));
	for (j=0;j<(int)camera->h;j++) {
		for (i=0;i<nThreads;i++) {
			memcpy(buff, info[i].outputBuffer + 3*j*frameInc, 3*frameInc*sizeof(float));
			buff += 3*frameInc;
		}
	}

	// free outputBuffers first
	for (i=0;i<nThreads;i++) {
		free(info[i].outputBuffer);
	}

	// free thread info
	free(threadIDs);
	free(threadHandles);
	free(info);
}

#else
// unable to match platform-specific threading, so calling a multi-threaded function will just
// get you the single-threaded functions.
void (*eosRenderScene_STANDARD_MULTITHREADED) (eosScene *scene, eosCamera *camera, float *outputBuffer, const eosRendererSettings &settings) = eosRenderScene_STANDARD_SINGLETHREADED;
void (*eosRenderScene_SUBSAMPLE_MULTITHREADED) (eosScene *scene, eosCamera *camera, float *outputBuffer, const eosRendererSettings &settings) = eosRenderScene_SUBSAMPLE_SINGLETHREADED;
#endif

/// <summary>Takes scene information and chooses the appropriate rendering function based on the settings flags. Scene is rendered
/// into outputBuffer which should be a (float *) with (camera->w*camera->h*3) elements. User is responsible for allocating before rendering.</summary>
/// <returns>(void)<returns>
void eosRenderScene(eosScene *scene, eosCamera *camera, float *outputBuffer, const eosRendererSettings &settings) {
	// first, cycle through the scene and prepare all the lights
	int i;
	for (i=0;i<scene->nLights;i++) {
		eosPrepareLight(scene->lights[i]);
	}

	if (!ao_init) {
		eosInitAORandTable();
	}

	if (settings.flags & EOS_RENDERER_STANDARD) {
		if (settings.flags & EOS_RENDERER_MULTITHREADED) {
			eosRenderScene_STANDARD_MULTITHREADED(scene, camera, outputBuffer, settings);
		} else {
			eosRenderScene_STANDARD_SINGLETHREADED(scene, camera, outputBuffer, settings);
		}
	} else if (settings.flags & EOS_RENDERER_SUBSAMPLE) {
		if (settings.flags & EOS_RENDERER_MULTITHREADED) {
			eosRenderScene_SUBSAMPLE_MULTITHREADED(scene, camera, outputBuffer, settings);
		} else {
			eosRenderScene_SUBSAMPLE_SINGLETHREADED(scene, camera, outputBuffer, settings);
		}
	}
}
