/*
Eos Scene - definitions
2013 - Matt Pierce

	Defines the functions declared in eosScene.h

TAB WIDTH: 4
*/

#include <stdio.h>
#include <string.h>				// for memcpy
#include "eosRendererMesh.h"
#include "eosLight.h"
#include "eosEnvironmentMap.h"
#include "eosScene.h"

/// <summary>Returns an initialized scene. Use this when you create a new
/// scene to ensure you don't pass uninitialized pointers to any of the Clear functions</summary>
/// <returns>An initialized (empty) scene</returns>
eosScene eosInitScene() {
	eosScene tmp;
	
	tmp.nLights = tmp.nMeshes = 0;
	tmp.capLights = tmp.capMeshes = EOS_SCENE_ALLOCATION_INCREMENT;

	tmp.env = NULL;
	tmp.lights = (eosLight**)malloc(tmp.capLights*sizeof(eosLight**));
	tmp.meshes = (eosRendererMesh**)malloc(tmp.capMeshes*sizeof(eosRendererMesh**));

	return tmp;
}

/// <summary>Adds a light to the scene.</summary>
/// <returns>(void)</returns>
void eosAddLight(eosScene *scene, eosLight *light) {
	eosLight **tmpBuffer;

	// if adding 1 light would put us over capacity
	if (scene->nLights+1 >= scene->capLights) {
		scene->capLights += EOS_SCENE_ALLOCATION_INCREMENT;						// increase capacity
		tmpBuffer = (eosLight**)malloc(scene->capLights*sizeof(eosLight*));		// allocate a new buffer
		memset(tmpBuffer, 0, scene->capLights*sizeof(eosLight*));				// set all pointers to NULL (maybe remove for speed?)
		memcpy(tmpBuffer, scene->lights, scene->nLights*sizeof(eosLight*));		// copy all existing lights
		free(scene->lights);													// free old buffer
		scene->lights = tmpBuffer;												// make the new buffer our current buffer
	}
	scene->lights[scene->nLights] = light;		// add light
	scene->nLights++;							// increment count of lights
}

/// <summary>Removes all lights from the scene</summary>
/// <returns>(void)</returns>
void eosClearLights(eosScene *scene) {
	free(scene->lights);
	scene->nLights = 0;
	scene->capLights = EOS_SCENE_ALLOCATION_INCREMENT;
	scene->lights = (eosLight**)malloc(scene->capLights*sizeof(eosLight*));
}

/// <summary>Adds a mesh to the scene</summary>
/// <returns>(void)</returns>
void eosAddMesh(eosScene *scene, eosRendererMesh *mesh) {
	eosRendererMesh **tmpBuffer;

	// if adding 1 mesh would put us over capacity
	if (scene->nMeshes+1 >= scene->capMeshes) {
		scene->capMeshes += EOS_SCENE_ALLOCATION_INCREMENT;									// increase capacity
		tmpBuffer = (eosRendererMesh**)malloc(scene->capMeshes*sizeof(eosRendererMesh*));	// allocate a new buffer
		memset(tmpBuffer, 0, scene->capMeshes*sizeof(eosRendererMesh*));					// set all pointers to NULL
		memcpy(tmpBuffer, scene->meshes, scene->nMeshes*sizeof(eosRendererMesh*));			// copy all existing meshes
		free(scene->meshes);																// free old buffer
		scene->meshes = tmpBuffer;															// make the new buffer our current buffer
	}
	scene->meshes[scene->nMeshes] = mesh;
	scene->nMeshes++;
}

/// <summary>Removes all meshes from the scene</summary>
/// <returns>(void)</returns>
void eosClearMeshes(eosScene *scene) {
	free(scene->meshes);
	scene->nMeshes = 0;
	scene->capMeshes = EOS_SCENE_ALLOCATION_INCREMENT;
	scene->meshes = (eosRendererMesh **)malloc(scene->capMeshes*sizeof(eosRendererMesh*));
}

/// <summary>Adds an environment map to the scene. Unlike lights and meshes
/// you can only have 1 environment map in your scene</summary>
/// <returns>(void)</returns>
void eosAddEnvMap(eosScene *scene, eosEnvironmentMap *env) {
	scene->env = env;
}

/// <summary>Removes the environment map from the scene.</summary>
/// <returns>(void)</returns>
void eosClearEnvMap(eosScene *scene) {
	scene->env = NULL;
}

/// <summary>Removes all lights, meshes, and environment maps at once.</summary>
/// <returns>(void)</returns>
void eosClearScene(eosScene *scene) {
	eosClearLights(scene);
	eosClearMeshes(scene);
	eosClearEnvMap(scene);
}

/// <summary>Traces all the meshes in the scene, with the ability to use
/// capT to avoid tracing trees and triangles that are beyond initial contacts.</summary>
/// <returns>trace result</returns>
eosTraceResult eosTraceScene(eosScene *scene, 
							const float &px, const float &py, const float &pz, 
							const float &dx, const float &dy, const float &dz, 
							const float &capT) {

	int i;								// loop counter
	eosTraceResult trBest;				// best result so far
	eosTraceResult trTemp;				// for catching returned results
	
	trBest.t = -1;						// so far, no contact

	// loop through each mesh
	for (i=0;i<scene->nMeshes;i++) {
		// trace the mesh, capped at our best hit so far
		if (capT<0) {
			trTemp = eosTraceMesh(scene->meshes[i], px, py, pz, dx, dy, dz, trBest.t>=0?trBest.t:-1);
		} else {
			trTemp = eosTraceMesh(scene->meshes[i], px, py, pz, dx, dy, dz, (trBest.t>=0 && trBest.t<capT)?trBest.t:capT);
		}
		// if the result is better than our best, make it our new best
		if ((trBest.t<0) || ((trTemp.t >= 0) && (trTemp.t<trBest.t))) {
			trBest = trTemp;
			trBest.mesh = i;
		}
	}

	return trBest;
}
