/*
Eos Texture - definitions
2013 - Matt Pierce

	Definitions for the functions described in eosTexture.h

TAB WIDTH: 4
*/

#include <stdio.h>			// oddly enough, for NULL identifier
#include "eosTexture.h"

/// <summary>A shortcut function for creating a texture.</summary>
/// <returns>a newly created texture</returns>
eosTexture eosCreateTexture(int width, int height, unsigned char planes, 
							unsigned char type, unsigned char filter, 
							unsigned char repeatX, unsigned char repeatY, 
							void * buff) {
	eosTexture temp;

	// set buff to NULL - don't set this until very last
	// if you get a texture back with a null buffer it indicates that
	// something went wrong
	temp.buff = (void *)NULL;

	if (width<0) return temp;
	temp.w = width;
	while (width>0) {
		if ((width & 0x01) && (width>1)) return temp;
		width = width >> 1;
	}

	if (height<0) return temp;
	temp.h = height;
	while (height>0) {
		if ((height & 0x01) && (height>1)) return temp;
		height = height >> 1;
	}

	temp.planes = planes;
	if (planes<1 || planes>4) return temp;

	temp.type = type;
	if (type!=EOS_TEXTURE_BYTE && type!=EOS_TEXTURE_FLOAT) return temp;

	temp.filter = filter;
	if (filter!=EOS_TEXTURE_NEAREST && type!=EOS_TEXTURE_BILINEAR) return temp;

	temp.repeatX = repeatX;
	if (repeatX!=EOS_TEXTURE_CLAMP && repeatX!=EOS_TEXTURE_REPEAT) return temp;

	temp.repeatY = repeatY;
	if (repeatY!=EOS_TEXTURE_CLAMP && repeatY!=EOS_TEXTURE_REPEAT) return temp;

	// We got through all our sanity checks OK, so attach the
	// buffer and return. USER IS RESPONSIBLE FOR FREEING THIS BUFFER
	temp.buff = buff;
	return temp;
}

inline int eosMax(const int &a, const int &b) {return a<b?b:a;}
inline int eosMin(const int &a, const int &b) {return a>b?b:a;}

/// <summary>Samples the color for coords (s,t) for a texture with options
/// type=EOS_TEXTURE_BYTE and filter=EOS_TEXTURE_NEAREST. Result placed in output</summary>
/// <returns>(void)</returns>
void eosSampleTexture_BYTE_NEAREST(const eosTexture &tex, float s, float t, float *output) {
	int tx = (int)(s*tex.w);
	int ty = (int)(t*tex.h);
	unsigned char *buff = (unsigned char *)tex.buff;

	tx = (tex.repeatX==EOS_TEXTURE_CLAMP)?
		eosMax(0, eosMin(tex.w-1, tx)):
		tx & (tex.w-1);
	ty = (tex.repeatY==EOS_TEXTURE_CLAMP)?
		eosMax(0, eosMin(tex.h-1, ty)):
		ty & (tex.h-1);
	ty = tex.h-ty-1;
	buff += tex.planes * (ty*tex.w+tx);
	switch (tex.planes) {
		case 1: 
			output[0] = output[1] = output[2] = buff[0]/255.0f;
			output[3] = 1.0f;
			break;
		case 2: 
			output[0] = output[1] = output[2] = buff[0]/255.0f;
			output[3] = buff[1]/255.0f;
			break;
		case 3: 
			output[0] = buff[0]/255.0f;
			output[1] = buff[1]/255.0f;
			output[2] = buff[2]/255.0f;
			output[3] = 1.0f;
			break;
		case 4: 
			output[0] = buff[0]/255.0f;
			output[1] = buff[1]/255.0f;
			output[2] = buff[2]/255.0f;
			output[3] = buff[3]/255.0f;
			break;
		default: 
			break;
	}
}

/// <summary>Samples the color for coords (s,t) for a texture with options
/// type=EOS_TEXTURE_FLOAT and filter=EOS_TEXTURE_NEAREST. Result placed in output</summary>
/// <returns>(void)</returns>
void eosSampleTexture_FLOAT_NEAREST(const eosTexture &tex, float s, float t, float *output) {
	int tx = (int)(s*tex.w);
	int ty = (int)(t*tex.h);
	float *buff = (float *)tex.buff;

	tx = (tex.repeatX==EOS_TEXTURE_CLAMP)?
		eosMax(0, eosMin(tex.w-1, tx)):
		tx & (tex.w-1);
	ty = (tex.repeatY==EOS_TEXTURE_CLAMP)?
		eosMax(0, eosMin(tex.h-1, ty)):
		ty & (tex.h-1);
	ty = tex.h-ty-1;
	buff += tex.planes * (ty*tex.w+tx);
	switch (tex.planes) {
		case 1: 
			output[0] = output[1] = output[2] = buff[0];
			output[3] = 1.0f;
			break;
		case 2: 
			output[0] = output[1] = output[2] = buff[0];
			output[3] = buff[1];
			break;
		case 3: 
			output[0] = buff[0];
			output[1] = buff[1];
			output[2] = buff[2];
			output[3] = 1.0f;
			break;
		case 4: 
			output[0] = buff[0];
			output[1] = buff[1];
			output[2] = buff[2];
			output[3] = buff[3];
			break;
		default: 
			break;
	}
}

void eosSampleTexture_FLOAT_BILINEAR(const eosTexture &tex, float s, float t, float *output) {
	float *buff1 = (float *)tex.buff;
	float *buff2 = buff1;
	float *buff3 = buff1;
	float *buff4 = buff1;

	// get top left corner
	int tx1 = (int)(s*tex.w);
	int ty1 = (int)(t*tex.h);
	// get pct
	float pctX = s*tex.w - tx1;
	float pctY = t*tex.h - ty1;
	float ipctX = 1.0f - pctX;
	float ipctY = 1.0f - pctY;
	// get other 3 corners
	int tx2 = tx1+1;
	int ty2 = ty1;
	int tx3 = tx1;
	int ty3 = ty1+1;
	int tx4 = tx1+1;
	int ty4 = ty1+1;

	// bound corners
	tx1 = (tex.repeatX==EOS_TEXTURE_CLAMP)?
		eosMax(0, eosMin(tex.w-1, tx1)):
		tx1 & (tex.w-1);
	ty1 = (tex.repeatY==EOS_TEXTURE_CLAMP)?
		eosMax(0, eosMin(tex.h-1, ty1)):
		ty1 & (tex.h-1);
	ty1 = tex.h-ty1-1;

	tx2 = (tex.repeatX==EOS_TEXTURE_CLAMP)?
		eosMax(0, eosMin(tex.w-1, tx2)):
		tx2 & (tex.w-1);
	ty2 = (tex.repeatY==EOS_TEXTURE_CLAMP)?
		eosMax(0, eosMin(tex.h-1, ty2)):
		ty2 & (tex.h-1);
	ty2 = tex.h-ty2-1;

	tx3 = (tex.repeatX==EOS_TEXTURE_CLAMP)?
		eosMax(0, eosMin(tex.w-1, tx3)):
		tx3 & (tex.w-1);
	ty3 = (tex.repeatY==EOS_TEXTURE_CLAMP)?
		eosMax(0, eosMin(tex.h-1, ty3)):
		ty3 & (tex.h-1);
	ty3 = tex.h-ty3-1;

	tx4 = (tex.repeatX==EOS_TEXTURE_CLAMP)?
		eosMax(0, eosMin(tex.w-1, tx4)):
		tx4 & (tex.w-1);
	ty4 = (tex.repeatY==EOS_TEXTURE_CLAMP)?
		eosMax(0, eosMin(tex.h-1, ty4)):
		ty4 & (tex.h-1);
	ty4 = tex.h-ty4-1;

	buff1 += tex.planes * (ty1*tex.w+tx1);
	buff2 += tex.planes * (ty2*tex.w+tx2);
	buff3 += tex.planes * (ty3*tex.w+tx3);
	buff4 += tex.planes * (ty4*tex.w+tx4);

	switch (tex.planes) {
	case 1:
		output[0] = output[1] = output[2] = ipctY*(ipctX*buff1[0] + pctX*buff2[0]) + pctY*(ipctX*buff3[0]+pctX*buff4[0]);
		output[3] = 1.0f;
		break;
	case 2:
		output[0] = output[1] = output[2] = ipctY*(ipctX*buff1[0] + pctX*buff2[0]) + pctY*(ipctX*buff3[0]+pctX*buff4[0]);
		output[3] = ipctY*(ipctX*buff1[1] + pctX*buff2[1]) + pctY*(ipctX*buff3[1]+pctX*buff4[1]);
		break;
	case 3:
		output[0] = ipctY*(ipctX*buff1[0] + pctX*buff2[0]) + pctY*(ipctX*buff3[0]+pctX*buff4[0]);
		output[1] = ipctY*(ipctX*buff1[1] + pctX*buff2[1]) + pctY*(ipctX*buff3[1]+pctX*buff4[1]);
		output[2] = ipctY*(ipctX*buff1[2] + pctX*buff2[2]) + pctY*(ipctX*buff3[2]+pctX*buff4[2]);
		output[3] = 1.0f;
	case 4:
		output[0] = ipctY*(ipctX*buff1[0] + pctX*buff2[0]) + pctY*(ipctX*buff3[0]+pctX*buff4[0]);
		output[1] = ipctY*(ipctX*buff1[1] + pctX*buff2[1]) + pctY*(ipctX*buff3[1]+pctX*buff4[1]);
		output[2] = ipctY*(ipctX*buff1[2] + pctX*buff2[2]) + pctY*(ipctX*buff3[2]+pctX*buff4[2]);
		output[3] = ipctY*(ipctX*buff1[3] + pctX*buff2[3]) + pctY*(ipctX*buff3[3]+pctX*buff4[3]);
	default:
		break;
	}

}

void eosSampleTexture_BYTE_BILINEAR(const eosTexture &tex, float s, float t, float *output) {
	unsigned char *buff1 = (unsigned char *)tex.buff;
	unsigned char *buff2 = buff1;
	unsigned char *buff3 = buff1;
	unsigned char *buff4 = buff1;

	// get top left corner
	int tx1 = (int)(s*tex.w);
	int ty1 = (int)(t*tex.h);
	// get pct
	float pctX = s*tex.w - tx1;
	float pctY = t*tex.h - ty1;
	float ipctX = 1.0f - pctX;
	float ipctY = 1.0f - pctY;
	// get other 3 corners
	int tx2 = tx1+1;
	int ty2 = ty1;
	int tx3 = tx1;
	int ty3 = ty1+1;
	int tx4 = tx1+1;
	int ty4 = ty1+1;

	// bound corners
	tx1 = (tex.repeatX==EOS_TEXTURE_CLAMP)?
		eosMax(0, eosMin(tex.w-1, tx1)):
		tx1 & (tex.w-1);
	ty1 = (tex.repeatY==EOS_TEXTURE_CLAMP)?
		eosMax(0, eosMin(tex.h-1, ty1)):
		ty1 & (tex.h-1);
	ty1 = tex.h-ty1-1;

	tx2 = (tex.repeatX==EOS_TEXTURE_CLAMP)?
		eosMax(0, eosMin(tex.w-1, tx2)):
		tx2 & (tex.w-1);
	ty2 = (tex.repeatY==EOS_TEXTURE_CLAMP)?
		eosMax(0, eosMin(tex.h-1, ty2)):
		ty2 & (tex.h-1);
	ty2 = tex.h-ty2-1;

	tx3 = (tex.repeatX==EOS_TEXTURE_CLAMP)?
		eosMax(0, eosMin(tex.w-1, tx3)):
		tx3 & (tex.w-1);
	ty3 = (tex.repeatY==EOS_TEXTURE_CLAMP)?
		eosMax(0, eosMin(tex.h-1, ty3)):
		ty3 & (tex.h-1);
	ty3 = tex.h-ty3-1;

	tx4 = (tex.repeatX==EOS_TEXTURE_CLAMP)?
		eosMax(0, eosMin(tex.w-1, tx4)):
		tx4 & (tex.w-1);
	ty4 = (tex.repeatY==EOS_TEXTURE_CLAMP)?
		eosMax(0, eosMin(tex.h-1, ty4)):
		ty4 & (tex.h-1);
	ty4 = tex.h-ty4-1;

	buff1 += tex.planes * (ty1*tex.w+tx1);
	buff2 += tex.planes * (ty2*tex.w+tx2);
	buff3 += tex.planes * (ty3*tex.w+tx3);
	buff4 += tex.planes * (ty4*tex.w+tx4);

	switch (tex.planes) {
	case 1:
		output[0] = output[1] = output[2] = (ipctY*(ipctX*buff1[0] + pctX*buff2[0]) + pctY*(ipctX*buff3[0]+pctX*buff4[0]))/255.0f;
		output[3] = 1.0f;
		break;
	case 2:
		output[0] = output[1] = output[2] = (ipctY*(ipctX*buff1[0] + pctX*buff2[0]) + pctY*(ipctX*buff3[0]+pctX*buff4[0]))/255.0f;
		output[3] = ipctY*(ipctX*buff1[1] + pctX*buff2[1]) + pctY*(ipctX*buff3[1]+pctX*buff4[1]);
		break;
	case 3:
		output[0] = (ipctY*(ipctX*buff1[0] + pctX*buff2[0]) + pctY*(ipctX*buff3[0]+pctX*buff4[0]))/255.0f;
		output[1] = (ipctY*(ipctX*buff1[1] + pctX*buff2[1]) + pctY*(ipctX*buff3[1]+pctX*buff4[1]))/255.0f;
		output[2] = (ipctY*(ipctX*buff1[2] + pctX*buff2[2]) + pctY*(ipctX*buff3[2]+pctX*buff4[2]))/255.0f;
		output[3] = 1.0f;
	case 4:
		output[0] = (ipctY*(ipctX*buff1[0] + pctX*buff2[0]) + pctY*(ipctX*buff3[0]+pctX*buff4[0]))/255.0f;
		output[1] = (ipctY*(ipctX*buff1[1] + pctX*buff2[1]) + pctY*(ipctX*buff3[1]+pctX*buff4[1]))/255.0f;
		output[2] = (ipctY*(ipctX*buff1[2] + pctX*buff2[2]) + pctY*(ipctX*buff3[2]+pctX*buff4[2]))/255.0f;
		output[3] = (ipctY*(ipctX*buff1[3] + pctX*buff2[3]) + pctY*(ipctX*buff3[3]+pctX*buff4[3]))/255.0f;
	default:
		break;
	}

}

/// <summary>Chooses the appropriate routine and passes data along to
/// that routine. Result placed in float *output. Regardless of input texture
/// type, output is expected to be pre-allocated float[4]</summary>
/// <returns>(void)</returns>
void eosSampleTexture(const eosTexture &tex, float s, float t, float *output) {
	if (tex.buff) {
		if (tex.type == EOS_TEXTURE_BYTE) {
			if (tex.filter == EOS_TEXTURE_NEAREST) {
				eosSampleTexture_BYTE_NEAREST(tex, s, t, output);
			} else if (tex.filter == EOS_TEXTURE_BILINEAR) {
				eosSampleTexture_BYTE_BILINEAR(tex, s, t, output);
			}
		} else if (tex.type == EOS_TEXTURE_FLOAT) {
			if (tex.filter == EOS_TEXTURE_NEAREST) {
				eosSampleTexture_FLOAT_NEAREST(tex, s, t, output);
			} else if (tex.filter == EOS_TEXTURE_BILINEAR) {
				eosSampleTexture_FLOAT_BILINEAR(tex, s, t, output);
			}
		}
	}
}