#include <stdio.h>
#include <string.h>
#include <math.h>
#include "VectorMath.h"
#include "eosTriangle.h"
#include "eosTree.h"

/// <summary>Checks to see if a point is in a bounding box. Done this way to short-circuit with returns</summary>
/// <returns>true/false</returns>
inline bool	eosPointInAABB(const eosAABB &box, const float &px, const float &py, const float &pz) {
	if ((px < box.min[0]) || (px > box.max[0])) return false;		// check x axis
	if ((py < box.min[1]) || (py > box.max[1])) return false;		// check y axis
	if ((pz < box.min[2]) || (pz > box.max[2])) return false;		// check z axis
	
	// by process of elimination, point must be inside the box
	return true;
}

/// <summary>Builds an individual tree node out of the triangles remaining in tris</summary>
/// <returns>an individual tree node (and all of its children)</returns>
eosTreeNode eosBuildTreeNode (eosTriangleBuildList &tris, const eosAABB &nodeBounds, eosBuildInfo *buildInfo, eosTriangle *masterTriList) {
	eosTreeNode			retVal;				// return value
	int					i, li, ui;			// loop counters
	eosAABB				centerBounds;		// range of center points (used to figure out which axis to split)
	float				centerBoundary;		// center line
	eosAABB				absBounds;			// a bounding box of the actual geometry
	eosAABB				lowerBounds;		// a bounding box to pass to the child nodes
	eosAABB				upperBounds;		// a bounding box to pass to the child nodes

	eosTriangleBuildList	lowerList;		// list for lower node (if we need to split)
	eosTriangleBuildList	upperList;		// list for upper node (if we need to split)

	upperBounds = lowerBounds = nodeBounds;

	// if there's only 1 triangle
	if (tris.length == 1) {
		// make a leaf node
		retVal.leafNode = true;
		retVal.triIndex = tris.tris[0];
		// make sure to clean up allocated storage
		free(tris.tris);
		// return leaf node
		return retVal;
	}

	// otherwise, it's definitely not a leaf node
	retVal.leafNode = false;

	// find the variation of center points in the list we're working with
	centerBounds.min[0] = buildInfo[tris.tris[0]].cp[0];
	centerBounds.max[0] = buildInfo[tris.tris[0]].cp[0];
	centerBounds.min[1] = buildInfo[tris.tris[0]].cp[1];
	centerBounds.max[1] = buildInfo[tris.tris[0]].cp[1];
	centerBounds.min[2] = buildInfo[tris.tris[0]].cp[2];
	centerBounds.max[2] = buildInfo[tris.tris[0]].cp[2];
	memcpy(&absBounds, &(buildInfo[tris.tris[0]].box), sizeof(eosAABB));
	for (i=1;i<tris.length;i++) {
		if (buildInfo[tris.tris[i]].cp[0] < centerBounds.min[0]) centerBounds.min[0] = buildInfo[tris.tris[i]].cp[0];
		if (buildInfo[tris.tris[i]].cp[1] < centerBounds.min[1]) centerBounds.min[1] = buildInfo[tris.tris[i]].cp[1];
		if (buildInfo[tris.tris[i]].cp[2] < centerBounds.min[2]) centerBounds.min[2] = buildInfo[tris.tris[i]].cp[2];
		if (buildInfo[tris.tris[i]].cp[0] > centerBounds.max[0]) centerBounds.max[0] = buildInfo[tris.tris[i]].cp[0];
		if (buildInfo[tris.tris[i]].cp[1] > centerBounds.max[1]) centerBounds.max[1] = buildInfo[tris.tris[i]].cp[1];
		if (buildInfo[tris.tris[i]].cp[2] > centerBounds.max[2]) centerBounds.max[2] = buildInfo[tris.tris[i]].cp[2];
		if (buildInfo[tris.tris[i]].box.min[0] < absBounds.min[0]) absBounds.min[0] = buildInfo[tris.tris[i]].box.min[0];
		if (buildInfo[tris.tris[i]].box.min[1] < absBounds.min[1]) absBounds.min[1] = buildInfo[tris.tris[i]].box.min[1];
		if (buildInfo[tris.tris[i]].box.min[2] < absBounds.min[2]) absBounds.min[2] = buildInfo[tris.tris[i]].box.min[2];
		if (buildInfo[tris.tris[i]].box.max[0] > absBounds.max[0]) absBounds.max[0] = buildInfo[tris.tris[i]].box.max[0];
		if (buildInfo[tris.tris[i]].box.max[1] > absBounds.max[1]) absBounds.max[1] = buildInfo[tris.tris[i]].box.max[1];
		if (buildInfo[tris.tris[i]].box.max[2] > absBounds.max[2]) absBounds.max[2] = buildInfo[tris.tris[i]].box.max[2];
	}

	// see if there's any way to split out some empty space
	for (int ca = 0; ca < 3; ca++) {
		if ((nodeBounds.max[ca]-nodeBounds.min[ca]>EOS_EMPTY_NODE_MINIMUM) && ((absBounds.min[ca]-nodeBounds.min[ca])/(nodeBounds.max[ca]-nodeBounds.min[ca]) > EOS_EMPTY_NODE_THRESHOLD)) {
			// free space in lower X
			retVal.splitAxis = ca;
			float smallDiff = (nodeBounds.max[ca] - nodeBounds.min[ca])*0.01f;
			retVal.lowerBound = absBounds.min[retVal.splitAxis]-smallDiff;
			retVal.upperBound = absBounds.min[retVal.splitAxis]+smallDiff;
			retVal.absMinimum = absBounds.min[retVal.splitAxis];
			retVal.absMaximum = absBounds.max[retVal.splitAxis];
	
			lowerList.length = 0;
			lowerList.tris = NULL;
			upperList.length = tris.length;
			upperList.tris = (unsigned int *)malloc(upperList.length*sizeof(unsigned int));
			memcpy(upperList.tris, tris.tris, upperList.length*sizeof(unsigned int));
			
			free(tris.tris);
			
			retVal.lowNode = NULL; // must be null so we don't try to free() later
			retVal.highNode = (eosTreeNode *)malloc(sizeof(eosTreeNode));
			
			upperBounds.min[retVal.splitAxis] = absBounds.min[retVal.splitAxis];
			*(retVal.highNode) = eosBuildTreeNode(upperList, upperBounds, buildInfo, masterTriList);
	
			return retVal;
		} else if ((nodeBounds.max[ca]-nodeBounds.min[ca]>EOS_EMPTY_NODE_MINIMUM) && ((absBounds.max[ca]-nodeBounds.min[ca])/(nodeBounds.max[ca]-nodeBounds.min[ca]) < EOS_EMPTY_NODE_THRESHOLD)) {
			// free space in upper X
			retVal.splitAxis = ca;
			float smallDiff = (nodeBounds.max[ca] - nodeBounds.min[ca])*0.01f;
			retVal.lowerBound = absBounds.max[retVal.splitAxis]-smallDiff;
			retVal.upperBound = absBounds.max[retVal.splitAxis]+smallDiff;
			retVal.absMinimum = absBounds.min[retVal.splitAxis];
			retVal.absMaximum = absBounds.max[retVal.splitAxis];
	
			lowerList.length = tris.length;
			lowerList.tris = (unsigned int *)malloc(lowerList.length*sizeof(unsigned int));
			upperList.length = 0;
			upperList.tris = NULL;
			memcpy(lowerList.tris, tris.tris, lowerList.length*sizeof(unsigned int));
			
			free(tris.tris);
			
			retVal.lowNode = (eosTreeNode *)malloc(sizeof(eosTreeNode));
			retVal.highNode = NULL; // must be NULL so we don't try to free() later
			
			lowerBounds.max[retVal.splitAxis] = absBounds.max[retVal.splitAxis];
			*(retVal.lowNode) = eosBuildTreeNode(lowerList, lowerBounds, buildInfo, masterTriList);
	
			return retVal;
		}
	}

	// pick our split axis based on the axis with the greatest variation
	if ((centerBounds.max[0]-centerBounds.min[0])>(centerBounds.max[1]-centerBounds.min[1])) {
		if ((centerBounds.max[0]-centerBounds.min[0])>(centerBounds.max[2]-centerBounds.min[2])) {
			retVal.splitAxis = 0;
		} else {
			retVal.splitAxis = 2;
		}
	} else {
		if ((centerBounds.max[1]-centerBounds.min[1])>(centerBounds.max[2]-centerBounds.min[2])) {
			retVal.splitAxis = 1;
		} else {
			retVal.splitAxis = 2;
		}
	}

	// calculate our split value
	centerBoundary = 0.5f*(centerBounds.max[retVal.splitAxis] + centerBounds.min[retVal.splitAxis]);

	// count up the number of triangles on each side of the line
	lowerList.length = 0;
	upperList.length = 0;
	for (i=0;i<tris.length;i++) {
		if (buildInfo[tris.tris[i]].cp[retVal.splitAxis] <= centerBoundary) lowerList.length++;
		else upperList.length++;
	}

	// allocate storage
	lowerList.tris = (unsigned int *)malloc(lowerList.length*sizeof(unsigned int));
	upperList.tris = (unsigned int *)malloc(upperList.length*sizeof(unsigned int));

	// pick some boundaries that we know for certain will be overridden
	retVal.lowerBound = centerBoundary;
	retVal.upperBound = centerBoundary;
	retVal.absMinimum = absBounds.min[retVal.splitAxis];
	retVal.absMaximum = absBounds.max[retVal.splitAxis];

	// copy indices, and find bounds
	ui = 0;			// upper i
	li = 0;			// lower i
	for (i=0;i<tris.length;i++) {
		if (buildInfo[tris.tris[i]].cp[retVal.splitAxis] <= centerBoundary) {
			lowerList.tris[li] = tris.tris[i];
			li++;
			if (retVal.upperBound < buildInfo[tris.tris[i]].box.max[retVal.splitAxis]) {
				retVal.upperBound = buildInfo[tris.tris[i]].box.max[retVal.splitAxis];
			}
		} else {
			upperList.tris[ui] = tris.tris[i];
			ui++;
			if (retVal.lowerBound > buildInfo[tris.tris[i]].box.min[retVal.splitAxis]) {
				retVal.lowerBound = buildInfo[tris.tris[i]].box.min[retVal.splitAxis];
			}
		}
	}

	// free tris.tris NOW (doing before recursion gives us O(n) memory consumption, after O(n log(n)) or worse
	free(tris.tris);

	// allocate nodes
	retVal.lowNode = (eosTreeNode *)malloc(sizeof(eosTreeNode));
	retVal.highNode = (eosTreeNode *)malloc(sizeof(eosTreeNode));

	// set our child bounds
	lowerBounds.max[retVal.splitAxis] = retVal.upperBound;
	upperBounds.min[retVal.splitAxis] = retVal.lowerBound;

	// recurse
	*(retVal.lowNode)  = eosBuildTreeNode(lowerList, lowerBounds, buildInfo, masterTriList);
	*(retVal.highNode) = eosBuildTreeNode(upperList, upperBounds, buildInfo, masterTriList);

	return retVal;
}

inline float eosMin(const float &a, const float &b) {return a<b?a:b;}
inline float eosMax(const float &a, const float &b) {return a>b?a:b;}

/// <summary>Builds a full tree - calculates some tree-wide info, then kicks off building the root node</summary>
/// <returns>(void)</returns>
void eosBuildTree (eosTree *tree, eosTriangle *masterTriList, int nTris) {
	eosBuildInfo *				buildInfo;			// information about each triangle used for building the tree
	eosTriangleBuildList		buildList;			// list of triangles to pass to first node
													// (root node gets all triangles)
	int							i;					// loop counters

	// set some values in our tree
	tree->box.max[0] = tree->box.min[0] = masterTriList[0].p1[0];
	tree->box.max[1] = tree->box.min[1] = masterTriList[0].p1[1];
	tree->box.max[2] = tree->box.min[2] = masterTriList[0].p1[2];
	tree->tris = masterTriList;
	tree->nTris = nTris;
	tree->matrix.LoadIdentity();

	// Precalculate information used in the tree build process
	buildInfo = (eosBuildInfo *)malloc(nTris*sizeof(eosBuildInfo));
	for (i=0;i<nTris;i++) {
		// calculate center point
		buildInfo[i].cp[0] = (masterTriList[i].p1[0]+masterTriList[i].p2[0]+masterTriList[i].p3[0])/3.0f;
		buildInfo[i].cp[1] = (masterTriList[i].p1[1]+masterTriList[i].p2[1]+masterTriList[i].p3[1])/3.0f;
		buildInfo[i].cp[2] = (masterTriList[i].p1[2]+masterTriList[i].p2[2]+masterTriList[i].p3[2])/3.0f;
		// calculate bounds
		buildInfo[i].box.min[0] = eosMin(masterTriList[i].p1[0], eosMin(masterTriList[i].p2[0], masterTriList[i].p3[0]));
		buildInfo[i].box.min[1] = eosMin(masterTriList[i].p1[1], eosMin(masterTriList[i].p2[1], masterTriList[i].p3[1]));
		buildInfo[i].box.min[2] = eosMin(masterTriList[i].p1[2], eosMin(masterTriList[i].p2[2], masterTriList[i].p3[2]));
		buildInfo[i].box.max[0] = eosMax(masterTriList[i].p1[0], eosMax(masterTriList[i].p2[0], masterTriList[i].p3[0]));
		buildInfo[i].box.max[1] = eosMax(masterTriList[i].p1[1], eosMax(masterTriList[i].p2[1], masterTriList[i].p3[1]));
		buildInfo[i].box.max[2] = eosMax(masterTriList[i].p1[2], eosMax(masterTriList[i].p2[2], masterTriList[i].p3[2]));
		// while we're looping, update our full tree's bounding box
		tree->box.min[0] = eosMin(tree->box.min[0], buildInfo[i].box.min[0]);
		tree->box.min[1] = eosMin(tree->box.min[1], buildInfo[i].box.min[1]);
		tree->box.min[2] = eosMin(tree->box.min[2], buildInfo[i].box.min[2]);
		tree->box.max[0] = eosMax(tree->box.max[0], buildInfo[i].box.max[0]);
		tree->box.max[1] = eosMax(tree->box.max[1], buildInfo[i].box.max[1]);
		tree->box.max[2] = eosMax(tree->box.max[2], buildInfo[i].box.max[2]);
	}

	// make a list for our root node with all the triangles in it
	buildList.length = nTris;
	buildList.tris = (unsigned int *)malloc(nTris*sizeof(unsigned int));
	for (i=0;i<nTris;i++) {
		buildList.tris[i] = i;
	}

	// build our root node
	tree->rootNode = eosBuildTreeNode(buildList, tree->box, buildInfo, masterTriList);

	free(buildInfo);
}

/// <summary>Checks to see if a ray intersects a bounding box, and if so puts the
/// in and out points in t_near and t_far</summary>
/// <returns>true/false</returns>
bool eosRayAABBIntersect (eosAABB box, SVect3d p, SVect3d d, float &t_near, float &t_far) {
	bool			found_min_t = false;	// the minimum t for ray that's in the box
	bool			found_max_t = false;	// the maximum t for ray that's in the box
	SVect3d			v_near = v3d(0,0,0);	// used for finding out the "near" and "far" bounds
	SVect3d			v_far = v3d(0,0,0);		// of the bounding box (since any component of d can
											// be negative, min is not always near)

	if (d.x>0) {
		v_far.x = box.max[0];
		v_near.x = box.min[0];
	} else {
		v_far.x = box.min[0];
		v_near.x = box.max[0];
	}
	if (d.y>0) {
		v_far.y = box.max[1];
		v_near.y = box.min[1];
	} else {
		v_far.y = box.min[1];
		v_near.y = box.max[1];
	}
	if (d.z>0) {
		v_far.z = box.max[2];
		v_near.z = box.min[2];
	} else {
		v_far.z = box.min[2];
		v_near.z = box.max[2];
	}

	// finding t_far looks complicated, but it's only because we have
	// to be careful about tracking whether we've found anything or not
	// (since not every plane necessarily intersects the ray, and because
	// there's no invalid value)
	if (d.x!=0) {
		if (found_max_t) {
			t_far = eosMin(t_far, (v_far.x-p.x)/d.x);
		} else {
			t_far = (v_far.x-p.x)/d.x;
			found_max_t = true;
		}
	}
	if (d.y!=0) {
		if (found_max_t) {
			t_far = eosMin(t_far, (v_far.y-p.y)/d.y);
		} else {
			t_far = (v_far.y-p.y)/d.y;
			found_max_t = true;
		}
	}
	if (d.z!=0) {
		if (found_max_t) {
			t_far = eosMin(t_far, (v_far.z-p.z)/d.z);
		} else {
			t_far = (v_far.z-p.z)/d.z;
			found_max_t = true;
		}
	}

	// if p is inside the bounding box, then t_near is 0, and we're definitely valid
	if (eosPointInAABB(box, (float)p.x, (float)p.y, (float)p.z)) {
		t_near = 0;
		return true;
	// otherwise, find out where the ray enters the box
	} else {
		// finding t_near is way simpler since we can start at 0
		// and use max to pick the right one
		t_near = 0;
		if (d.x!=0) t_near = eosMax(t_near, (v_near.x-p.x)/d.x);
		if (d.y!=0) t_near = eosMax(t_near, (v_near.y-p.y)/d.y);
		if (d.z!=0) t_near = eosMax(t_near, (v_near.z-p.z)/d.z);

		if (t_far >= t_near) return true;
		else return false;
	}
}

/// <summary>Traces a ray from one tree node down (not recursive, but separating it keeps
/// the tree trace function cleaner)</summary>
/// <returns>trace result - t less than 0 indicates no contact</returns>
eosTraceResult eosTraceTreeNode (eosTreeNode rootNode,
										const float &px, const float &py, const float &pz,
										const float &dx, const float &dy, const float &dz,
										float tMin, float tMax,
										eosTriangle *masterTriangleList,
										float capT) {
	float				p[3], d[3], invd[3];				// vertices stored in arrays (to exploit axis selection)
	eosTraceResult		etrBest;							// our best result so far
	eosTraceResult		etrTemp;							// a temporary result to compare to our best

	eosTreeNode *		node = &rootNode;					// the node we're currently checking
	eosTreeNode *		nodeStack[EOS_KD_TREE_STACK_SIZE];	// a stack of nodes for tree traversal
	float				minStack[EOS_KD_TREE_STACK_SIZE];	// a stack of tMin values
	float				maxStack[EOS_KD_TREE_STACK_SIZE];	// a stack of tMax values
	int					stackPtr=0;							// stack pointer

	int					nTraces=0;							// number of traces (view in debugger)

	p[0] = px; p[1] = py; p[2] = pz;
	d[0] = dx; d[1] = dy; d[2] = dz;
	if (dx!=0) invd[0] = 1.0f/dx;
	if (dy!=0) invd[1] = 1.0f/dy;
	if (dz!=0) invd[2] = 1.0f/dz;

	etrBest.t = -1;		// default to no contact

	while (true) {
		if (node == NULL) {
			// I'm going to try introducing empty nodes. This hopefully will
			// vastly reduce the number of triangle traces per ray.

			// if you're tracing an empty node, do nothing, then pop the stack.
			if (stackPtr == 0) {
				return etrBest;
			} else {
				stackPtr--;
				tMin = minStack[stackPtr];
				tMax = maxStack[stackPtr];
				node = nodeStack[stackPtr];
			}
		} else if (node->leafNode) {
			// first make sure this node is still worth tracing
			// (if this was pushed before we got a contact, it may be possible
			// to rule it out after we get one)
			if (!(etrBest.t>=0 && etrBest.t<tMin)) {
				// trace the triangle in the leaf and see if it beats our best so far
				nTraces++;
				etrTemp = eosTestTrace(masterTriangleList[node->triIndex], px, py, pz, dx, dy, dz);
				if (etrTemp.t>=0 && (etrBest.t<0 || etrBest.t>etrTemp.t) && (capT<0.0f || etrTemp.t<capT)) {
					etrBest = etrTemp;
					etrBest.tri = node->triIndex;	// only bother copying value if we hit
				}
			}
			// since nodes can overlap we can't just return here, we have to pop and keep going
			// but we'll use etrBest.t instead of tMin to eliminate more nodes
			if (stackPtr == 0) {
				return etrBest;
			} else {
				stackPtr--;
				tMin = minStack[stackPtr];
				tMax = maxStack[stackPtr];
				node = nodeStack[stackPtr];
			}
		} else if (capT>0.0f && capT<tMin) {
			// this node starts after our cap, so pop the stack
			if (stackPtr == 0) {
				return etrBest;
			} else {
				stackPtr--;
				tMin = minStack[stackPtr];
				tMax = maxStack[stackPtr];
				node = nodeStack[stackPtr];
			}
		} else if (etrBest.t>=0 && etrBest.t<tMin) {
			// since we still have to check nodes after getting a contact due to overlap, check
			// here to see if we can eliminate this node and all of its children at least
			if (stackPtr == 0) {
				return etrBest;
			} else {
				stackPtr--;
				tMin = minStack[stackPtr];
				tMax = maxStack[stackPtr];
				node = nodeStack[stackPtr];
			}
		} else {				// not a leaf node
			bool posDir = true;
			bool zeroError = false;
			float lowerT, upperT;
			float absMinT, absMaxT;
			unsigned char sa = node->splitAxis;

			
			// find out where our ray intersects the split axis (with
			// a special case if it's co-planar to split)
			if (d[sa]<0) {
				posDir = false;
			} else if (d[sa]==0) {
				zeroError = true;
			}
			if (!zeroError) {
				lowerT = (node->lowerBound - p[sa])*invd[sa];
				upperT = (node->upperBound - p[sa])*invd[sa];
				absMinT = (node->absMinimum - p[sa])*invd[sa];
				absMaxT = (node->absMaximum - p[sa])*invd[sa];
			}

			// if the ray is co-planar
			if (zeroError) {
				if (p[sa] > node->upperBound) {
					// only hits the higher node
					node = node->highNode;
				} else if (p[sa] < node->lowerBound) {
					// only hits the lower node
					node = node->lowNode;
				} else {
					// hits both nodes (somewhere between lowerBound and upperBound)
					// tMin and tMax do not change on a zeroError case
					minStack[stackPtr] = tMin;
					maxStack[stackPtr] = tMax;
					nodeStack[stackPtr] = node->highNode;
					stackPtr++;
					node = node->lowNode;
				}
			} else if (posDir) {
				// if we're traveling from low to high ...
				if ((absMinT>tMax) || (absMaxT<tMin)) {
					// doesn't hit any geometry in this node, so pop the stack
					if (stackPtr == 0) {
						return etrBest;
					} else {
						stackPtr--;
						tMin = minStack[stackPtr];
						tMax = maxStack[stackPtr];
						node = nodeStack[stackPtr];
					}
				} else if (lowerT>tMax) {
					// only hits lower node (also, no need to edit tMin/tMax)
					node = node->lowNode;
				} else if (upperT<tMin) {
					// only hits upper node (also, no need to edit tMin/tMax)
					node = node->highNode;
				} else {
					// hits both nodes - push the highNode and trace the lowNode
					minStack[stackPtr] = eosMax(lowerT, tMin);
					maxStack[stackPtr] = tMax;
					nodeStack[stackPtr] = node->highNode;
					stackPtr++;
					tMax = eosMin(upperT, tMax);
					node = node->lowNode;
				}
			} else {
				// if we're traveling from high to low ...
				if ((absMaxT>tMax) || (absMinT<tMin)) {
					// doesn't hit any geometry in this node, so pop the stack
					if (stackPtr == 0) {
						return etrBest;
					} else {
						stackPtr--;
						tMin = minStack[stackPtr];
						tMax = maxStack[stackPtr];
						node = nodeStack[stackPtr];
					}
				} else if (lowerT<tMin) {
					// only hits lower node (also, no need to edit tMin/tMax)
					node = node->lowNode;
				} else if (upperT>tMax) {
					// only hits upper node (also, no need to edit tMin/tMax)
					node = node->highNode;
				} else {
					// hits both nodes - push the lowNode and trace the highNode
					minStack[stackPtr] = eosMax(upperT, tMin);
					maxStack[stackPtr] = tMax;
					nodeStack[stackPtr] = node->lowNode;
					stackPtr++;
					tMax = eosMin(lowerT, tMax);
					node = node->highNode;
				}
			}
		}
	}

	return etrBest;		// always return etrBest, even if it indicates no contact
}

/// <summary>Traces the tree - first finds out if/where the ray hits the bounding box around the tree,
/// then starts tracing at the rootNode. capT lets you limit how far you'll allow short-circuit out of
/// traces that have a limited range (like seeing if a ray hits a light)(-1 means no limit).</summary>
/// <returns>trace result - t less than 0 indicates no contact</returns>
eosTraceResult eosTraceTree (eosTree *tree, 
							const float &px, const float &py, const float &pz,
							const float &dx, const float &dy, const float &dz,
							float capT) {
	SVect3d				p = v3d(px, py, pz);
	SVect3d				d = v3d(dx, dy, dz);
	float				minT = 0;
	float				maxT = 0;
	eosTraceResult tr;

	// default result - no contact
	tr.t = -1;

	// transform ray from world space to object space
	p = tree->matrix.InverseTransform(p);
	d = tree->matrix.InverseTransform(d) - tree->matrix.InverseTransform(v3d(0,0,0));

	// if the ray intersects the tree's bounding box
	if (eosRayAABBIntersect(tree->box, p, d, minT, maxT)) {
		// we also get the entry/exit points from that function
		// in minT and maxT - we need those to trace the tree

		// if the minT is before the capT, don't even go to the nodes
		if (capT>=0 && minT>capT) {
			return tr;
		}
		return eosTraceTreeNode(tree->rootNode, p.x, p.y, p.z, d.x, d.y, d.z, minT, maxT, tree->tris, capT);
	} else {
		// if we don't hit the bounding box, we must not hit any triangles
		return tr;
	}
}

/// <summary>Destroys tree nodes allocated on the heap. Should not be called directly by user</summary>
/// <returns>(void)</returns>
void eosDestroyNode (eosTreeNode *node) {
	// NULL nodes are allowed, so always check before freeing
	if (node != NULL) {
		if (!node->leafNode) {
			eosDestroyNode(node->lowNode);
			eosDestroyNode(node->highNode);
		}
		free(node);
	}
}

/// <summary>Destroys all tree nodes in a tree (the root node is not on the heap, so this is a bit of a kludge)</summary>
/// <returns>(void)</returns>
void eosClearTree (eosTree *tree) {
	if (!tree->rootNode.leafNode) {
		eosDestroyNode(tree->rootNode.lowNode);
		eosDestroyNode(tree->rootNode.highNode);
	}
}

/// <summary>Call this to rebuild the tree if you altered some or all of the triangles. This is a little
/// costly so don't call this every frame for static geometry.</summary>
/// <returns>(void)</returns>
void eosRebuildTree (eosTree *tree, eosTriangle *masterTriList, int nTris) {
	eosClearTree(tree);
	eosBuildTree(tree, masterTriList, nTris);
}
