/*
Eos Tree - header
2013 - Matt Pierce

	Defines the Bounded KD Tree used to organize our triangles.
	The tree minimizes the number of triangle-ray intersection
	tests we need to do, so it's essential that this is done
	right.

TAB WIDTH: 4
*/

#ifndef _EOS_TREE_H
#define _EOS_TREE_H

#include "VectorMath.h"
#include "eosTriangle.h"

const int EOS_KD_TREE_STACK_SIZE = 256;		// the maximum depth of our tree, allows for 2^n triangles
											// 256 allows for nearly a googol triangles (more than we can store in memory)

const float EOS_EMPTY_NODE_THRESHOLD = 0.1f;	// If (n)% of the node is empty before you hit the triangles' bounding box
												// then split that box into an empty node and a node containing triangles

const float EOS_EMPTY_NODE_MINIMUM = 0.01F;		// The minimum size to allow division into an empty node.

// eosAABB - Axis-aligned bounding box
//		used to do some quick elimination of rays that don't even
//		hit the tree at all, much less any nodes or triangles
struct eosAABB {
	float				min[3];				// the minimum x,y,z of our box
	float				max[3];				// the maximum x,y,z of our box
};

// eosBuildInfo - information used when building the tree
//		This information is calcuated once for each triangle and
//		then passed along when building the tree
struct eosBuildInfo {
	eosAABB				box;				// a bounding box around the triangle
	float				cp[3];				// the triangle's center point (avg of 3 verts)
};

// eosTriangleBuildList - temporary list of triangles
//		This stores the list of triangles that need to be sorted
//		for any given node throughout the build process. The
//		indexes reference eosTree.tris
struct eosTriangleBuildList {
	int					length;				// the length of the list
	unsigned int *		tris;				// array of _indexes_ of triangles
};

// eosTreeNode - one node of the tree
//		Nodes either hold one triangle, or they're split into
//		two child nodes. Since this is a bounded KD tree, the
//		nodes have two overlapping boundaries. By allowing
//		two nodes to share some space, we don't have to split
//		triangles or place them in both nodes.
struct eosTreeNode {
	unsigned char		splitAxis;			// the axis on which this node is split
	bool				leafNode;			// if this is a leaf node, it contains a triangle
	unsigned int		triIndex;			// index of triangle (if leaf node)
	eosTreeNode *		lowNode;			// Node below upperBound (if not leaf node)
	eosTreeNode *		highNode;			// Node above lowerBound (if not leaf node)
	float				lowerBound;			// where the "far" node starts (if not leaf node)
	float				upperBound;			// where the "close" node ends (if not leaf node)
	float				absMinimum;			// the absolute minimum vertex
	float				absMaximum;			// the absolute maximum vertex
};

// eosTree - the tree itself
//		Stores the root node and triangles. A matrix is used
//		to transform incoming vectors (instead of rebuilding
//		the entire tree whenever things move around).
struct eosTree {
	eosTreeNode			rootNode;			// the root of our tree
	eosAABB				box;				// a bounding box (for quick eliminations)
	CMatrix				matrix;				// a matrix (for transforming tree w/o needing to rebuild)
	eosTriangle *		tris;				// the triangles associated with this tree
	unsigned int		nTris;				// the number of triangles in the tree
};


inline bool			eosPointInAABB		(const eosAABB &box, const float &px, const float &py, const float &pz);
eosTreeNode			eosBuildTreeNode	(eosTriangleBuildList &tris, const eosAABB &nodeBounds, eosBuildInfo *buildInfo, eosTriangle *masterTriList);
void				eosBuildTree		(eosTree *tree, eosTriangle *masterTriList, int nTris);
bool				eosRayAABBIntersect	(eosAABB box, SVect3d p, SVect3d d, float &t_near, float &t_far);
eosTraceResult		eosTraceTreeNode	(eosTreeNode rootNode,
										const float &px, const float &py, const float &pz,
										const float &dx, const float &dy, const float &dz,
										float tMin, float tMax,
										eosTriangle *masterTriangleList,
										float capT=-1.0f);
eosTraceResult		eosTraceTree		(eosTree *tree, 
										const float &px, const float &py, const float &pz,
										const float &dx, const float &dy, const float &dz,
										float capT=-1.0f);
void				eosDestroyNode		(eosTreeNode *node);
void				eosRebuildTree		(eosTree *tree, eosTriangle *masterTriList, int nTris);
void				eosClearTree		(eosTree *tree);
#endif // _EOS_TREE_H