/*
Eos Triangle - definitions
2013 - Matt Pierce

	Defines all the functions declared in eosTriangle.h

TAB WIDTH: 4
*/
#include "eosTriangle.h"
#include "VectorMath.h"

/// <summary>Pre-calculates some values before tracing.</summary>
/// <returns>(void)</returns>
void eosSetupTriangle(eosTriangle &tri) {
	float denom2;
	SVect3d temp;
	int u, v;

	// get the plane nomral as the cross product between two edges of the triangle
	temp = vmNormalize(vmCrossProduct(
		vmNormalize(v3d(tri.p2[0], tri.p2[1], tri.p2[2]) - v3d(tri.p1[0],tri.p1[1],tri.p1[2])),
		vmNormalize(v3d(tri.p3[0], tri.p3[1], tri.p3[2]) - v3d(tri.p1[0],tri.p1[1],tri.p1[2]))
	));
	// store the plane normal
	tri.pnx = (float)temp.x;
	tri.pny = (float)temp.y;
	tri.pnz = (float)temp.z;
	// last, calculate the plane offset
	tri.pd = -(tri.pnx*tri.p1[0] + tri.pny*tri.p1[1] + tri.pnz*tri.p1[2]);

	// Find the primary axis of the triangle
	if (fabs(tri.pnx)>fabs(tri.pny)) {
		if (fabs(tri.pnx)>fabs(tri.pnz)) {
			tri.PrimaryAxis = 0;
		} else {
			tri.PrimaryAxis = 2;
		}
	} else {
		if (fabs(tri.pny)>fabs(tri.pnz)) {
			tri.PrimaryAxis = 1;
		} else {
			tri.PrimaryAxis = 2;
		}
	}

	// u and v are temporary variables we use to select axes
	// in the following precalculation
	u = (tri.PrimaryAxis+1)%3;
	v = (tri.PrimaryAxis+2)%3;

	// precalculate everything we can precalcuate about the ray-triangle intersection test
	tri.bu = tri.p3[u] - tri.p1[u];
	tri.bv = tri.p3[v] - tri.p1[v];
	tri.cu = tri.p2[u] - tri.p1[u];
	tri.cv = tri.p2[v] - tri.p1[v];
	denom2 = -1.0f/(tri.bu*tri.cv - tri.bv*tri.cu);
	tri.bu *= denom2;
	tri.bv *= denom2;
	tri.cu *= denom2;
	tri.cv *= denom2;
}

/// <summary>Traces a ray at a triangle and tells you if there was a contact. Called
/// a test trace because the renderer still needs to do a bunch of work to get a
/// color out of this information.</summary>
/// <returns>eosTraceResult - the result of trace; t less than 0 indicates no contact</returns>
eosTraceResult eosTestTrace(const eosTriangle &tri,
		const float &px, const float &py, const float &pz,
		const float &dx, const float &dy, const float &dz) {

	float				t;				// the point along the ray at which we make contact
	float				cp[3];			// the point at which we make contact
	float				ku, kv;			// temporary variables used in the calcuation
	unsigned char		u, v;			// used with tri.PrimaryAxis to select planes
	eosTraceResult		tr;				// the variable we'll actually be returning

	// assume no contact
	tr.t = -1.0f;

	// first check to see if it hits the plane at t>=0
	t = tri.pnx*dx + tri.pny*dy + tri.pnz*dz;
	if (t==0) return tr;	// ray is coplanar
	t = -(tri.pnx*px + tri.pny*py + tri.pnz*pz + tri.pd)/t;
	if (t<0) return tr;		// ray intersection is behind ray origin

	// get contact point
	cp[0] = px + dx*t;
	cp[1] = py + dy*t;
	cp[2] = pz + dz*t;

	// get u/v
	u = (tri.PrimaryAxis+1)%3;
	v = (tri.PrimaryAxis+2)%3;

	// get ku/kv
	ku = cp[u] - tri.p1[u];
	kv = cp[v] - tri.p1[v];

	// get a1, a2, a3
	tr.a2 = (ku*tri.bv - kv*tri.bu);
	if (tr.a2<0) return tr;
	tr.a3 = (kv*tri.cu - ku*tri.cv);
	if (tr.a3<0) return tr;
	tr.a1 = 1.0f-tr.a2-tr.a3;
	if (tr.a1<0) return tr;

	// found contact, give t result
	tr.t = t;
	return tr;

}