/*
Eos Renderer Mesh - definitions
2013 - Matt Pierce

	Defines the functions in eosRendererMesh.h

TAB WIDTH: 4
*/

#include "eosRendererMesh.h"

/// <summary>Traces a ray toward a mesh and returns the result.</summary>
/// <returns>result from the trace</returns>
eosTraceResult eosTraceMesh(eosRendererMesh *mesh,
							const float &px, const float &py, const float &pz,
							const float &dx, const float &dy, const float &dz,
							float capT) {
	eosTraceResult tr;
	tr.t = -1.0f;

	SVect3d p = v3d(px, py, pz);
	SVect3d d = v3d(dx, dy, dz);

	p = mesh->matrix.InverseTransform(p);
	d = mesh->matrix.InverseTransform(d) - mesh->matrix.InverseTransform(v3d(0,0,0));

	tr = eosTraceTree(mesh->tree, p.x, p.y, p.z, d.x, d.y, d.z, capT);

	return tr;
}


/// <summary>Uses COBJFile to load an OBJ file, then copies data into a renderer mesh.</summary>
/// <returns>void - requires pointer to allocated mesh</returns>
void eosRendererMeshFromObj (const char *inFilename, eosRendererMesh* inMesh) {
	CObjFile obj;
	std::vector<CObjGroupPtr>::iterator groupIterator;
	std::vector<SObjMaterialPtr>::iterator materialIterator;
	std::vector<SObjTriPtr>::iterator triangleIterator;
	SObjMaterial* objMat;
	int i;

	obj.ReadFile(inFilename);

	// get material count
	inMesh->nMats = obj.list_mtl.size();

	// allocate materials
	inMesh->mats = (eosMaterial*)malloc(inMesh->nMats*sizeof(eosMaterial));

	// copy materials
	i = 0;
	for (materialIterator=obj.list_mtl.begin(); materialIterator!=obj.list_mtl.end(); ++materialIterator) {
		objMat = (*materialIterator);
		inMesh->mats[i] = eosCreateMaterial();
		inMesh->mats[i].color[0] = objMat->diffuse.x;
		inMesh->mats[i].color[1] = objMat->diffuse.y;
		inMesh->mats[i].color[2] = objMat->diffuse.z;
		inMesh->mats[i].color[3] = objMat->alpha;

		if (objMat->uses_texture) {
			// exploit the fact that the CObjFile sets a 1024 char limit
			inMesh->mats[i].texFile = (char *)malloc(1024*sizeof(char));
			memcpy((void *)inMesh->mats[i].texFile, (void *)objMat->sTextureName, 1024*sizeof(char));
			// once you're done loading from OBJ file, it's YOUR responsibility to handle file loading
			// using whatever library you see fit to use.
		}

		inMesh->mats[i].glossReflection = objMat->shininess;
		inMesh->mats[i].specular = vmLength(objMat->specular);
		inMesh->mats[i].fresnelTerm = objMat->fresnel_term;
		inMesh->mats[i].unshaded = !(objMat->lit);

		++i;
	}

	// iterate through CObjFile groups, get polygon count
	inMesh->nTris = 0;
	for (groupIterator=obj.list_g.begin(); groupIterator!=obj.list_g.end(); ++groupIterator) {
		inMesh->nTris += (*groupIterator)->tri.size();
	}

	// allocate triangles
	inMesh->tris = (eosTriangle *)malloc(inMesh->nTris*sizeof(eosTriangle));
	inMesh->renderInfo = (eosTriangleRenderInfo *)malloc(inMesh->nTris*sizeof(eosTriangleRenderInfo));

	i = 0;
	// iterate through CObjFile groups
	for (groupIterator=obj.list_g.begin(); groupIterator!=obj.list_g.end(); ++groupIterator) {
		// iterate through triangles
		for (triangleIterator=(*groupIterator)->tri.begin(); triangleIterator!=(*groupIterator)->tri.end(); ++triangleIterator) {
			// copy triangles
			inMesh->tris[i].p1[0] = obj.list_v[(*triangleIterator)->p1]->x;
			inMesh->tris[i].p1[1] = obj.list_v[(*triangleIterator)->p1]->y;
			inMesh->tris[i].p1[2] = obj.list_v[(*triangleIterator)->p1]->z;
			inMesh->tris[i].p2[0] = obj.list_v[(*triangleIterator)->p2]->x;
			inMesh->tris[i].p2[1] = obj.list_v[(*triangleIterator)->p2]->y;
			inMesh->tris[i].p2[2] = obj.list_v[(*triangleIterator)->p2]->z;
			inMesh->tris[i].p3[0] = obj.list_v[(*triangleIterator)->p3]->x;
			inMesh->tris[i].p3[1] = obj.list_v[(*triangleIterator)->p3]->y;
			inMesh->tris[i].p3[2] = obj.list_v[(*triangleIterator)->p3]->z;

			eosSetupTriangle(inMesh->tris[i]);

			inMesh->renderInfo[i].mat = &(inMesh->mats[(*triangleIterator)->mtl]);

			inMesh->renderInfo[i].n1[0] = obj.list_vn[(*triangleIterator)->n1]->x;
			inMesh->renderInfo[i].n1[1] = obj.list_vn[(*triangleIterator)->n1]->y;
			inMesh->renderInfo[i].n1[2] = obj.list_vn[(*triangleIterator)->n1]->z;
			inMesh->renderInfo[i].n2[0] = obj.list_vn[(*triangleIterator)->n2]->x;
			inMesh->renderInfo[i].n2[1] = obj.list_vn[(*triangleIterator)->n2]->y;
			inMesh->renderInfo[i].n2[2] = obj.list_vn[(*triangleIterator)->n2]->z;
			inMesh->renderInfo[i].n3[0] = obj.list_vn[(*triangleIterator)->n3]->x;
			inMesh->renderInfo[i].n3[1] = obj.list_vn[(*triangleIterator)->n3]->y;
			inMesh->renderInfo[i].n3[2] = obj.list_vn[(*triangleIterator)->n3]->z;

			if (inMesh->renderInfo[i].mat->texFile) {
				inMesh->renderInfo[i].tc1[0] = obj.list_vt[(*triangleIterator)->tc1]->x;
				inMesh->renderInfo[i].tc1[1] = obj.list_vt[(*triangleIterator)->tc1]->y;
				inMesh->renderInfo[i].tc2[0] = obj.list_vt[(*triangleIterator)->tc2]->x;
				inMesh->renderInfo[i].tc2[1] = obj.list_vt[(*triangleIterator)->tc2]->y;
				inMesh->renderInfo[i].tc3[0] = obj.list_vt[(*triangleIterator)->tc3]->x;
				inMesh->renderInfo[i].tc3[1] = obj.list_vt[(*triangleIterator)->tc3]->y;
			} else {
				inMesh->renderInfo[i].tc1[0] = inMesh->renderInfo[i].tc1[1] = 0.0f;
				inMesh->renderInfo[i].tc2[0] = inMesh->renderInfo[i].tc2[1] = 0.0f;
				inMesh->renderInfo[i].tc3[0] = inMesh->renderInfo[i].tc3[1] = 0.0f;
			}

			inMesh->renderInfo[i].tracerGroup = 1;

			++i;
		}
	}

	// Allocate and Build Tree
	inMesh->tree = (eosTree *)malloc(sizeof(eosTree));
	eosBuildTree(inMesh->tree, inMesh->tris, inMesh->nTris);

	// Initialize Matrix
	inMesh->matrix.LoadIdentity();

	// Paint texture groups
}